# coding: utf-8
from __future__ import with_statement
import caty
import caty.jsontools as json
from caty.jsontools import xjson
from caty.core.schema.base import JsonSchemaError
from caty.jsontools import *
from caty.core.command import Builtin, Internal
from caty.core.command.exception import *
from caty.core.exception import *
from caty.core.handler import AppDispatcher, RequestHandler
from caty.core.template import Template, PlainFile
from caty.core.facility import PEND
from caty.template.core.renderer import *
from caty.command import MafsMixin

import os
import sys
import traceback
import codecs
from optparse import Values
from StringIO import StringIO
from threading import RLock


schema = u"""
type list<_T> = [_T*];
type tuple<_S, _T> = [_S, _T];

type CatyExecMode = list <string>;

type DirectoryEntry = {"name":string, "isDir":boolean, *:any};

type ErrorObj = {
    "isError": boolean,
    "orig": string,
    "val": any,
    "message": string,
    "errorId": string?,
    "param": {*:any}?,
    *: any
};

type pinteger = integer(minimum=0);

type ErrorItem = ErrorObj & {
    "name": string,
    *: any
};

type ErrorList = list <ErrorItem>;

type ErrorMap = {
    *: ErrorObj
};

type ErrorInfo = {
    "errorList": ErrorList,
    "errorMap": ErrorMap
};

type MiniErrorInfo = MiniErrorObj | MiniError;

type MiniErrorObj = {
    *: string
};

type MiniError = string;

type WebInput = FormInput | JsonInput;

// 本当は @form とすべきだろうが使い勝手を考えて
type FormInput = {
    *: [WebItem, WebItem*]
};

type JsonInput = @json any;

type Response = {
    "status": integer,
    "header": {
        "content-type": string?,
        "content-length": integer?,
        *: string
    },
    "body": string | binary,
    "encoding": string? // テキストデータを返すときは必ずこれを追加
};

type Redirect = {
    "status": string,
    "header": {
        "Location": string,
        *: string
    }
};

type ErrorStatus = {
    "status": string
};

type AppInfo = {
    "name": string,
    "description": string,
    "path": string,
    *:any
};

type AppManifest = {
    "implLang": string,
    "description": string,
    "assign": {
        *:string,
    },
    *:any
};

type Assoc = {
    // プロパティ名はverb名、デフォルトのverb名（無名）は""
    *:string
};

type Assocs = {
    *: Assoc
};

type FileType = {
    "isText": boolean,
    "contentType": string,
    "description": string?
};

type FileTypes = {
    *: FileType
};

type UploadFile = {
    "filename": string,
    "data": binary
};

// Catyの例外データは次の型のサブタイプでなくてはならない
// 実装言語の例外を直接投げることも禁止しない（禁止できない）

/** 人間可読なメッセージ */
type Message = {
  // プロパティ名は言語識別子, ja, en など
  * : string?
} (minProperties = 1);

type CatyException = @* {
 "message" : Message, 

 // 以下の2つのプロパティは、必要なら分類に使用する
 "class" : string?,
 "id" : (integer|string)?,

 // 実装言語のスタックトレースデータ
 "stackTrace" : any?,
 
  *: any?
};

/**
 * マニフェストにおけるアプリケーション固有情報
 */
type CatyAppProperty = {
    "$typename": string,
    "$callback": string?,
    *: any
};

/**
 * 組み込み型の名前
 */
type BuiltinTypeName = (
   "number" | "string" | "boolean" | "null" | "binary" // スカラー型
 | "array" | "object" // 複合型
 | "tagged"  // XJSONによる拡張
 | "unknown" // 不明
);


"""

from pbc import PbcObject, Contract
   
class Expand(Builtin, MafsMixin, PbcObject):
    command_decl = u"""
    /**
     * 簡易的な print コマンド。
     *
     * print と同様の働きをするが、出力が HTTP 応答ヘッダのない単なる文字列となる。
     * コマンドラインシェルから起動する事を前提に作られている。
     * バイナリデータを読み込んだ場合、特に加工せずそのまま出力する。
     *
     */
    command expand {"raw": boolean, "resolve": boolean, "mode":string} [string] :: 
        any -> string | binary
        uses [pub, env, include, interpreter]
        refers python:caty.command.builtin.Expand;
    """
    __properties__ = ['path']

    def setup(self, opts, path):
        self.path, self.fs = self.parse_canonical_path(path)
        self.raw = opts.raw
        self.resolve = opts.resolve
        self.mode = opts.mode
        PbcObject.__init__(self)

    def execute(self, input):
        return self._expand(input)

    def dup(self, path):
        if not path.startswith('include@') and not '@' in path:
            path = 'include@' + path
        p = Expand(['--resolve'], [path])
        p.pub = self.pub
        p.include = self.include
        p.env = self.env
        p.interpreter = self.interpreter
        p._prepare()
        return p

    def _load_context(self, path, input):
        u"""path に対応した caty スクリプトファイルを実行する。
        """
        builder = self.interpreter
        assoc_path =[path+ '.icaty', path + '.caty']
        for p in assoc_path:
            if self.fs.open(p, 'rb').exists:
                content = self.fs.open(p).read()
                cmd = builder.build(content)
                return cmd(input)
        return None

    def subtemplate(self, path, input):
        sub = self.dup(path)
        sub.builder = self.interpreter
        return sub._expand(input)
 
    def _expand(self, input):
        path = self.path
        _f = self.fs.open(path)
        if not _f.exists:
            raise CatyException(
                'HTTP_404', 
                {
                    'ja': u'ファイルが見つかりません: %s' % path,
                    'en': u'File not found: %s' % path
                })
        if _f.is_dir:
            raise CatyException(
                'HTTP_403',
                {
                    'ja': u'403 Forbidden: %s' % path,
                    'en': u'403 Forbidden: %s' % path,
                }
            )
        if not self.env.exists('FILE_PATH'):
            self.env.put('FILE_PATH', path)
        assoc = False
        if self.raw:
            fo = self.fs.open(path)
            return fo.read()
        if self.resolve:
            assoc = True
        try:
            debug = self.env.get('DEBUG') if self.env.exists('DEBUG') else False
            template = Template(path, assoc, self.fs, debug)
            template.set_include_callback(self.subtemplate)
            if assoc:
                o = self._load_context(path, input)
                if o != None:
                    input = o
            if self.mode:
                m = opts.mode
                if m == 'text':
                    template.renderer = TextRenderer()
                elif m == 'html':
                    pass # デフォルトの動作
            fo = self.fs.open(path)
            template.context = input
            template.context['_FILE_PATH'] = self.env.get('FILE_PATH')
            template.context['_APP_PATH'] = self.env.get('APP_PATH')
            template.context['_HOST_URL'] = self.env.get('HOST_URL')
            template.context['_CATY_VERSION'] = self.env.get('CATY_VERSION')
            encoding = fo.encoding if fo.encoding else self.env.get('APP_ENCODING')
            is_text = fo.is_text
            outfile = codecs.getwriter(encoding)(StringIO())
            template.write(outfile)
            outfile.seek(0)
            if is_text:
                return unicode(outfile.read(), encoding)
            else:
                return outfile.read()
        except PlainFile:
            self.raw = True
            return self._expand(input)

    if caty.DEBUG:
        def __invariant__(self):
            assert self.path != None
            assert self.fs != None

        def _not_property_changed(self, r, old, *args, **kwds):
            assert self.path == old.path

        execute = Contract(execute)
        execute.ensure += _not_property_changed

class Response(Builtin):
    command_decl = u"""
    /**
     * 入力値を HTTP の応答ボディとし、 HTTP 応答ヘッダを付けて返す。
     * --ext オプションは拡張子の指定であり、このオプションが指定された場合は content-type ヘッダが対応した値となる。
     * 指定されなかった場合は application/octet-stream が用いられる。
     */
    command response {"ext":string, 
                      "status": integer, 
                      "content-type": string, 
                      "encoding": string} :: string | binary -> Response
        reads [env, pub]
        refers python:caty.command.builtin.Response;
    """
    def setup(self, opts):
       self.ext = opts.ext
       self.content_type = opts.content_type
       self.encoding = opts.encoding
       self.status = opts.status if opts.status else 200

    def execute(self, input):
        if not self.encoding:
            encoding = self.env.get('APP_ENCODING')
        else:
            encoding = self.encoding
        if self.content_type:
            tp = self.content_type
            if self.encoding:
                tp += '; charset=%s' % encoding
        elif self.ext:
            tp = self.pub.get_type(self.ext)
            if self.pub.is_text_ext(self.ext):
                tp += '; charset=%s' % encoding
        else:
            tp = 'application/octet-stream'
        if isinstance(input, unicode):
            length = len(input.encode(encoding))
        else:
            length = len(input)
        return {
            'status': self.status,
            'encoding': encoding,
            'body': input,
            'header': {
                'content-type': unicode(tp),
                'content-length': length
            }
        }

class Print(Expand):
    u"""組み込みの print コマンド。
    """
    command_decl = u"""
    /**
     * 入力をテンプレートに対して適用し、 HTTP 応答ヘッダを付けて返す。
     *
     * --raw オプションが指定された場合、テンプレートを評価せずにそのまま返す。
     *
     * --resolve オプションが指定された場合、インライン/アウトオブラインスクリプトをテンプレートの入力として扱う。
     * 両者が同時に指定された場合、インラインスクリプトが優先して使用される。
     *
     * --mode オプションは text か html のどちらかを値としてとる。
     * text が指定された場合、 HTML 特殊文字の自動エスケープが働かなくなる。
     * デフォルトでは html が指定された状態であり、 Web ページの出力にはこちらを使う。
     * text はメールの文面の出力など、 Web ページの出力以外の場合に用いること。
     */
    command print {"raw": boolean, 
                   "resolve": boolean, 
                   "mode":string} [string] :: 
        any -> Response
        uses [pub, include, interpreter, env]
        reads [schema]
        refers python:caty.command.builtin.Print;
    """
    def execute(self, input):
        path = self.path
        v = self._expand(input)
        fo = self.fs.open(path)
        if fo.is_text:
            ctype = '%s; charset=%s' % (fo.content_type, fo.encoding)
        else:
            ctype = fo.content_type
        encoding = fo.encoding if fo.encoding else self.env.get('APP_ENCODING')
        if isinstance(v, unicode):
            length = len(v.encode(encoding))
        else:
            length = len(v)
        return {
            'status': 200,
            'body': v,
            'header': {
                'content-length': length,
                'content-type': unicode(ctype)
            },
            'encoding': encoding
        }
 
class Found(Builtin):
    command_decl = u"""
    command found [string] :: void -> Response
        refers python:caty.command.builtin.Found;
    """

    def setup(self, path):
        self._path = path

    def execute(self):

        return {
            'status': 302,
            'body': u'',
            'header': {
                'Location': self._path
            }
        }

class Request(Internal):
    command_decl = u"""
    command request {
                    "verb": string, 
                    "method": string,
                    "debug": boolean,
                    "content-type": string} [string, string?] :: T -> T
        reads [env, schema]
        uses interpreter
        refers python:caty.command.builtin.Request;
    """
    def setup(self, opts, path, params=''):
        self._path = path
        self._method = opts.method or u'GET'
        self._opts = self._split_opts(params)
        if opts.verb:
            self._opts['_verb'] = opts.verb
        self._content_type = opts.content_type
        self._debug = opts.debug

    def _split_opts(self, params):
        r = {}
        for i in params.split('&'):
            if '=' in i:
                k, v = i.split('=')
                r[k] = v
        return r

    def execute(self, input):
        if self._content_type:
            content_type = self._content_type
        else:
            if not input:
                content_type = u''
            elif isinstance(input, dict):
                content_type = 'application/x-www-form-urlencoded'
            elif isinstance(input, TaggedValue):
                content_type = 'application/json'
            else:
                content_type = 'application/octed-stream'
        dispatcher = AppDispatcher(self._system)
        app = dispatcher.dispatch(self._path)
        f = app.create_facilities(lambda: self._facilities['session'])
        app.init_env(f, caty.DEBUG, ['web'], self._system, {'CONTENT_TYPE': content_type})
        for k, v in f.items():
            v.merge_transaction(self._facilities[k])
        handler = RequestHandler(f['interpreter'], 
                                 f['env'],
                                 app.pub, 
                                 app.web_config, 
                                 app.encoding)
        cmd = handler.request(self._path, self._opts, self._method, PEND)
        cmd.exec_cmd(input, self._debug)
        if cmd.succeed:
            for k, v in self._facilities.items():
                v.merge_transaction(f[k])
        return cmd.result

from topdown import *
from caty.core.schema.base import TagSchema, UnionSchema
from caty.core.casm.language import util, schemaparser
import operator
class TypeCalculator(object):
    def set_schema(self, s):
        scm = self.parse(s)
        self.convertor = scm

    def parse(self, s):
        return as_parser(self.parse_type).run(s, auto_remove_ws=True)

    def parse_type(self, seq):
        return chainl(self.term, self.op)(seq)

    def term(self, seq):
        return seq.parse([self.type, self.tag, self.group])
    
    def tag(self, seq):
        _ = seq.parse('@')
        n = seq.parse(self.tag_name)
        t = seq.parse(self.type_name)
        return TagSchema(n, self.schema[n])

    def group(self, seq):
        _ = seq.parse('(')
        d = seq.parse(self.parse_type)
        _ = seq.parse(')')
        return d

    def op(self, seq):
        o = seq.parse(['&', '|', '<<'])
        if o == '&':
            return operator.and_
        elif o == '<<':
            return operator.lshift
        else:
            return operator.or_

    def name(self, seq):
        return util.name(seq)
    
    def type(self, seq):
        return self.schema[self.type_name(seq)]

    def type_name(self, seq):
        return schemaparser.typename(seq)

    def error_report(self, e):
        if hasattr(e, 'to_path'):
            x = e.to_path()
            r = {}
            for k, v in x.items():
                r[unicode(k)] = v['message'] if isinstance(v['message'], unicode) else unicode(str(v['message']))
            return tagged('NG', r)
        else:
            return tagged('NG', unicode(e.message, 'utf-8') if isinstance(e.message, str) else e.message)

class Translate(Builtin, TypeCalculator):
    command_decl = u"""
        /**
         * 引数で指定された型に入力値を変換して返す。
         */
        command translate [string] :: WebInput -> @OK _T | @NG MiniErrorInfo
            reads schema
            refers python:caty.command.builtin.Translate;
    """
    def setup(self, schema_name):
        self.schema_name = schema_name
        self.set_schema(schema_name)

    def execute(self, input):
        try:
            t, v = split_tag(input)
            scm = self.convertor
            if t == 'form' or t == 'object':
                return tagged('OK', scm.convert(self._to_nested(v)))
            elif t == 'json':
                try:
                    scm.validate(v)
                    return tagged('OK', v)
                except JsonSchemaError, e:
                    return self.error_report(e)
        except JsonSchemaError, e:
            return self.error_report(e)

    def _to_nested(self, input):
        d = {}
        for k, v in input.items():
            d['$.'+k] = v
        return path2obj(d) if d else d

class Validate(Builtin, TypeCalculator):
    command_decl = u"""
        /**
         * 引数で指定された型で入力値を検証する。
         */
        command validate {"pred":boolean} [string] :: any -> @OK _T | @NG MiniErrorInfo | boolean
            reads schema
            refers python:caty.command.builtin.Validate;
    """
    def setup(self, opts, schema_name):
        self.pred = opts.pred
        self.schema_name = schema_name
        self.set_schema(schema_name)

    def execute(self, input):
        try:
            scm = self.convertor
            scm.validate(input)
            if self.pred:
                return True
            else:
                return tagged('OK', input)
        except JsonSchemaError, e:
            if self.pred:
                return False
            return self.error_report(e)

def fo_to_json(fo, long_format):
    if not long_format:
        return {u'name': fo.basename, u'isDir': fo.is_dir}
    else:
        return {
            u'name': fo.basename, 
            u'abspath': fo.path,
            u'isDir': fo.is_dir,
            u'lastModified': unicode(fo.last_modified) if not fo.is_dir else u''
            }

def namecmp(a, b):
    u"""FileObjectのソート用の比較関数。
    パス名を比較に用いる。
    """
    return cmp(a.path, b.path)

def has_ext(fo, ext):
    return fo.path.endswith(ext)

def kind_of(ent, kind):
    if kind == "file":
        return not ent.is_dir
    elif kind == "dir":
        return ent.is_dir
    else:
        return True

class Void(Builtin):
    command_decl = u"""
        /**
         * 入力を捨てるコマンド。
         * 
         */
        command void :: _T -> void
          refers python:caty.command.builtin.Void;
    """
    def execute(self, input):
        return None

class Version(Builtin):
    command_decl = u"""
        /**
         * バージョン情報の表示。
         *
         */
        command version :: void -> string 
          reads env
          refers python:caty.command.builtin.Version;
    """
    def execute(self):
        return self.env.get('CATY_VERSION')

class LsDir(Builtin):
    command_decl = u"""
        /**
         * ディレクトリ一覧の表示。
         *
         * 第一引数で指定されたディレクトリを読み込み、配下のファイルとディレクトリ一覧を返す。
         * 第二引数は拡張子の指定であり、例えば .html が指定された場合 .html 拡張子のファイル一覧だけが返される。
         * 通常はファイル名とファイルかディレクトリの種別のみが買えされるが、
         * --long オプションが指定された場合は最終更新時刻なども返される。
         */
        command lsdir {"long":boolean, "kind":string} [string, string?] :: 
            void -> list<DirectoryEntry>
            reads pub
            refers python:caty.command.builtin.LsDir;
    """
    def setup(self, opts, path, ext=''):
        self.path = path
        self.ext = ext
        self.long_format = opts.long
        kind = opts.kind
        self.kind = "file" if kind == "file" else ("dir" if kind == "dir"  else "any")

    def execute(self):
        from caty.util import bind2nd
        dirobj = self.pub.DirectoryObject(self.path)
        files = list(sorted([ent for ent in dirobj.read() 
                             if has_ext(ent, self.ext) and kind_of(ent, self.kind)],
                            cmp=namecmp))
        r = map(bind2nd(fo_to_json, self.long_format), files)
        return r

class PassData(Builtin):
    command_decl = u"""
        /**
         * 入力をそのまま出力にコピーする。
         *
         */
        command pass :: _T -> _T
            refers python:caty.command.builtin.PassData;
    """
    
    def execute(self, input):
        return input

class Nth(Builtin):
    command_decl = u"""
        /**
         * 配列の n 番目のデータを返す。 item との違いはこちらは 1 始まりということである。
         */
        command nth [integer] :: [any*] -> _T
            refers python:caty.command.builtin.Nth;
    """
    def setup(self, n):
        self.num = n

    def execute(self, input):
        n = self.num
        if n < 1:
            raise IndexError(n)
        x = input[n - 1] # 1始まり
        return x

class Item(Builtin):
    command_decl = u"""
        /**
         * 配列の添字 n のデータを返す。 nth との違いはこちらは 0 始まりということである。
         */
        command item [integer] :: [any*] -> _T
            refers python:caty.command.builtin.Item;
    """
    def setup(self, n):
        self.num = n

    def execute(self, input):
        x = input[self.num] # 0 はじまり
        return x

class GetPV(Builtin):
    command_decl = u"""
        /**
         * JSON オブジェクトから引数で与えられた名前のプロパティの値を取得する。
         * 値が存在しなかった場合は実行時エラーとなる。
         */
        command pv [string] :: object -> _T
            refers python:caty.command.builtin.GetPV;
    """
    def setup(self, key):
        self.key = key

    def execute(self, input):
        x = input[self.key]
        return x

class FindPV(Builtin):
    command_decl = u"""
    /**
     * JSON オブジェクトから引数で与えられた名前のプロパティの値を取得する。
     * 値が存在した場合 @EXISTS タグがついた値が返り、そうでなければ @NO タグのついた null が返る。
     */
    command findpv [string] :: object -> @EXISTS _T | @NO null
            refers python:caty.command.builtin.FindPV;
    """
    def setup(self, key):
        self.key = key

    def execute(self, input):
        if self.key in input:
            return tagged('EXISTS', input[self.key])
        else:
            return tagged('NO', None)

class Equals(Builtin):
    command_decl = u"""
        /**
         * 二つの入力値の比較を行う。
         * 両者が同じであれば @Same タグ付きで値が返り、そうでなければ @Diff タグ付きで返る。
         *
         */
        command eq :: tuple<any, any> -> @Same any | @Diff tuple<any, any>
            refers python:caty.command.builtin.Equals;
    """
    def execute(self, input):
        if input[0] == input[1]:
            return tagged('Same', input[0])
        else:
            return tagged('Diff', input)

class Eval(Builtin):
    command_decl = u"""
        /**
         * 入力値二つのうち第一要素を Caty スクリプトの式、第二要素をその入力として実行する。
         * このコマンドはテストモードでしか動作しないため、通常は使う必要はない。
         */
        @[test] command eval :: tuple<string, any> | string -> _T
            uses interpreter
            refers python:caty.command.builtin.Eval;
    """
    def execute(self, input):
        if isinstance(input, unicode):
            cmd = self.interpreter.build(input, transaction=PEND)
            return cmd(None)
        else:
            cmd = self.interpreter.build(input[0], transaction=PEND)
            return cmd(input[1])
        

class GetTag(Builtin):
    command_decl = u"""
        /**
         * 入力値のタグを取得する。組み込み型が与えられた場合、その型名がタグとして扱われる。
         */
        command tag :: _T -> string
            refers python:caty.command.builtin.GetTag;
    """
    def execute(self, input):
        return tag(input)

class Tagged(Builtin):
    command_decl = u"""
        /**
         * 第一の入力値をタグ名とし、第二の入力値をタグ付きにして返す。
         */
        command tagged :: tuple <string, _S> -> _T
            refers python:caty.command.builtin.Tagged;
    """
    def execute(self, input):
        return tagged(input[0], input[1])

class Untagged(Builtin):
    command_decl = u"""
        /**
         * 入力値のタグを除去して返す。組み込み型が与えられた場合、その値がそのまま返る。
         */
        command untagged [string?] :: S -> T
            refers python:caty.command.builtin.Untagged;
    """
    def setup(self, t=None):
        self.__tag = t

    def execute(self, input):
        if self.__tag is not None:
            t = tag(input)
            if t != self.__tag:
                raise Exception(u'タグ %s ではありません: %s' % (self.__tag, t))
        return untagged(input)

class ConsoleOut(Builtin):
    command_decl = u"""
        /**
         * 標準出力に入力値を書き出す。
         */
        @[console] command cout :: string -> void
            updates stream
            refers python:caty.command.builtin.ConsoleOut;
    """
    def execute(self, input):
        self.stream.write(input)
        self.stream.write('\n')

class ConsoleIn(Builtin):
    command_decl = u"""
        /**
         * 標準入力より値を読み込み、それを JSON オブジェクトに変換して返す。
         */
        @[console] command cin :: void -> any
            reads stream
            refers python:caty.command.builtin.ConsoleIn;
    """
    def execute(self):
        return xjson.loads(self.stream.read())

class DisplayApp(Builtin):
    command_decl = u"""
        /**
         * 現在のアプリケーションのアプリケーション情報を返す。
         */
        command app :: void -> AppInfo
            updates stream
            reads env
            refers python:caty.command.builtin.DisplayApp;
    """
    def execute(self):
        return self.env.get('CATY_APP')

class DisplayApps(Builtin):
    command_decl = u"""
        /**
         * すべてのアプリケーションのアプリケーション情報を返す。
         */
        command apps :: void -> list<AppInfo>
            reads env
            refers python:caty.command.builtin.DisplayApps;
    """
    def execute(self):
        return self.env.get('CATY_APPS')

class Home(Builtin):
    command_decl = u"""
    /**
     * 現在の作業ディレクトリ名を返す。
     * --long オプションが指定された場合、フルパスでの表示となる。
     */
    @[console] command home {"long":boolean?} :: void -> string
        reads env
        refers python:caty.command.builtin.Home;
    """
    def setup(self, opts):
        self.full_path = opts.long

    def execute(self):
        if self.full_path:
            return self.env.get('CATY_HOME')
        else:
            return os.path.basename(self.env.get('CATY_HOME'))


class Manifest(Builtin):
    command_decl = u"""
    /**
     * 現在のアプリケーションのマニフェストを返す。
     * これは通常 _manifest.json に書かれている情報である。
     * _manifest.json が存在しない場合、 Caty が自動生成した設定が表示される。
     */
    @[console] command manifest :: void -> AppManifest
        reads env
        refers python:caty.command.builtin.Manifest;
    """
    def execute(self):
        return self.env.get('APP_MANIFEST')

class ExecContext(Print):
    command_decl = u"""
    /**
     * ファイルに関連付けられたスクリプトを実行する。
     *
     */
    @[console] command exec-context [string] :: _S -> _T
        uses [pub, include, interpreter]
        refers python:caty.command.builtin.ExecContext;
    """
    def setup(self, path):
        self.path = path

    def execute(self, input):
        path = self.path
        if path.startswith('/'):
            fs = self.pub
        else:
            path = '/' + path
            fs = self.include
        template = Template(path, True, fs, True)
        return self._load_context(path, fs, input)

class Assoc(Builtin):
    command_decl = u"""
    /**
     * ファイルに対する関連付けを出力する。
     *
     */
    @[console] command assoc  
                           :: void -> Assocs
                           {} [string]         :: void -> Assoc
                           {} [string, string] :: void -> string | null
        reads env
        refers python:caty.command.builtin.Assoc;
    """
    def setup(self, ext=None, verb=None):
        self.ext = ext
        self.verb = verb

    def execute(self):
        assoc = self.env.get('FILE_ASSOC')
        if self.ext is None:
            return assoc
        else:
            if self.verb is None:
                return assoc.get(self.ext, {})
            return assoc.get(self.ext, {}).get(self.verb, None)

class ShowFileType(Builtin):
    command_decl = u"""
    /**
     * ファイルタイプを出力する。
     *
     *
     */
    @[console] command filetype :: void -> FileTypes
                              {} [string] :: void -> FileType | null
            reads pub
            refers python:caty.command.builtin.ShowFileType;
    """
    def setup(self, ext=None):
        self.ext = ext

    def execute(self):
        if self.ext is None:
            return self.pub.all_types()
        else:
            return self.pub.all_types().get(self.ext, None)

class Forbidden(Builtin):
    command_decl = u"""
    /**
     * 403 Forbidden を返す。
     */
    command forbidden [string] :: void -> never
        refers python:caty.command.builtin.Forbidden;
    """
    def execute(self):
        self.exit({
            'status': 403})

class Param(Builtin):
    command_decl = u"""
    /**
     * 引数を出力にコピーする。
     * 引数を与えなかった場合、 null が返る。
     */
    command param [string] :: void -> string
                           :: void -> null
        refers python:caty.command.builtin.Param;
    """
    def setup(self, arg=None):
        self.arg = arg

    def execute(self):
        return self.arg

from caty.util.path import join
class Redirect(Builtin):
    command_decl = u"""
    /**
     * リダイレクト処理。引数のパスに対してそのままリダイレクトする。
     */
    command redirect [string] :: void -> never
                     :: string -> never
        breaks Redirect
        reads env
        refers python:caty.command.builtin.Redirect;
    """
    def setup(self, path=None):
        self.path = path

    def execute(self, input=None):
        if self.path is None:
            path = input
        else:
            path = self.path
        self.exit({
            'header': {
                'Location': unicode(join(self.env.get('HOST_URL'), path)),
            },
            'status': 302})

class LocalSchema(Builtin):
    command_decl = u"""
    /**
     * スクリプトローカルスキーマの定義。
     * CatyFIT で使うためのコマンドなので、ユーザが明示的に使う必要はないはずである。
     */
    command define-local-schema :: string -> void
        reads schema
        refers python:caty.command.builtin.LocalSchema;
    """
    def execute(self, input):
        self.schema.add_local(input)

class Binary(Builtin):
    command_decl = u"""
    /**
     * バイナリデータへの変換。
     * 引数でフォーマットを指定し、入力文字列をそのフォーマットのバイナリと解釈して返す。
     * 特に指定しなかった場合、 base64 と解釈される。
     */
    command binary [("base64"|"raw")?] :: string -> binary
        refers python:caty.command.builtin.Binary;
    """

    def setup(self, format='base64'):
        self.format = format

    def execute(self, input):
        if self.format == 'raw':
            return str(input)
        elif self.format == 'base64':
            import base64
            return base64.b64decode(input)

class ListEnv(Builtin):
    command_decl = u"""
    /**
     * 環境変数の一覧の出力
     */
    command env :: void -> object
        reads env
        refers python:caty.command.builtin.ListEnv;
    """

    def execute(self):
        r = {}
        for k, v in self.env.items():
            r[k] = v
        return r

class Properties(Builtin):
    command_decl = u"""
    /**
     * JSON オブジェクトのプロパティ名のリストを返す。
     */
    command properties :: object -> list<string>
        refers python:caty.command.builtin.Properties;
    """

    def execute(self, input):
        return map(unicode, input.keys())


from types import *
import decimal

def _type_of(input):
    input_type = type(input)
    if input_type == NoneType:
        type_name = u'null'
    elif input_type == BooleanType:
        type_name = u'boolean'
    elif input_type == IntType or input_type == LongType or input_type == FloatType or \
            isinstance(input, decimal.Decimal):
        type_name = u'number'
    elif input_type == UnicodeType:
        type_name = u'string'
    elif input_type == ListType:
        type_name = u'array'
    elif input_type == DictType:
        type_name = u'object'
    elif isinstance(input, caty.jsontools.TaggedValue):
        type_name = u'tagged'
    elif input_type == StringType: # Uuuum by HIYAMA
        type_name = u'binary'
    else:
        type_name = u'unknown'
    return type_name

class TypeOf(Builtin):
    command_decl = u"""
    /** 入力の型を判断する */
    command typeof :: any -> BuiltinTypeName 
        refers python:caty.command.builtin.TypeOf;
    """
    def execute(self, input):
        return _type_of(input)

class TypeIs(Builtin):
    command_decl = u"""
    /** 入力の型が引数で指定された名前の型であるかどうかを判断する */
    command typeis [BuiltinTypeName] :: any -> boolean
    refers python:caty.command.builtin.TypeIs;
    """
    def setup(self, type_name):
        self.type_name = type_name

    def execute(self, input):
        r = _type_of(input)
        return r == self.type_name

