#coding: utf-8
from caty.core.command import Builtin
from caty.fit.runner import FitRunner
from caty.jsontools import stdjson as json
name = u'fit'
schema = u"""
"""

class Run(Builtin):
    command_decl = u"""
    /**
     * CatyFIT の実行。
     * 引数のファイル名が指定された場合、単一のビヘイビアファイルを実行する。
     * 引数が指定されなかった場合、すべてのビヘイビアファイルを実行する。
     * force オプションが指定された場合、明示的に補足しなかった例外を補足し、
     * エラーではなく失敗として扱う。
     *
     * CatyFIT はデフォルトでは dry run でテストを行う。
     * 即ちあらゆるファシリティへの書き込み要求のトランザクションはテスト終了時に破棄される。
     * 実際にファイルシステムやデータベースへの書き込みを行ったテストを行いたい場合、
     * no-dry-run オプションを付けて実行すること。
     */
    command run 
             {"verbose": boolean, "force": boolean, "no-dry-run": boolean} :: void -> void
             {"verbose": boolean, "force": boolean, "no-dry-run": boolean} [string] :: void -> void
            reads [env, behaviors]
            uses [pub, interpreter, stream]
            refers python:caty.command.fit.Run;
    """
    def setup(self, opts, path=None):
        self._path = path
        self._verbose = opts.verbose
        self._force = opts.force
        self._no_dry_run = opts.no_dry_run

    def execute(self):
        if self._path:
            results = [self._test(self.behaviors.open(self._path))]
        else:
            results = []
            tests = self._list_test()
            for e in tests:
                results.append(self._test(e))
        for r in results: 
            # テストが終わる度にロールバックが行われるので、
            # 結果の書き込みは最後にまとめて行う
            r()

    def _list_test(self):
        d = self.behaviors.opendir('/')
        r = []
        for e in d.read(True):
            if e.path.endswith('.beh'):
                r.append(e)
        return r

    def _test(self, fo):
        if self._verbose:
            self.stream.write(u'%s を実行しています...\n' % fo.path)
        runner = FitRunner(fo, self.interpreter, self._force, self._no_dry_run)
        runner.run()
        report = runner.report
        #report['tests'] = self._list_summary(self.behaviors.opendir('/'))
        report['app'] = self._get_app_name()
        path = fo.path
        def _():
            if self._verbose:
                self.stream.write(u'%s の結果を書き込んでいます...\n' % fo.path)
            self._create_target_dir(path)
            self._write_result(report, path)
        return _

    def _create_target_dir(self, path):
        app = self._get_app_name()
        base_dir = 'fit-view:/%s' % app
        l = path.split('/')[:-1]
        dirs = ['/'.join(l[:i+1]) for i in range(len(l))]
        for d in dirs:
            target = self.pub.opendir(base_dir+d)
            if not target.exists:
                target.create()

    def _write_result(self, result, path):
        app = self._get_app_name()
        path = 'fit-view:/%s%s' % (app, path.replace('.beh', '.fit'))
        f = self.pub.open(path, 'wb')
        json.dump(result, f)
        f.close()

    def _get_app_name(self):
        return self.env.get('CATY_APP')['name']

    def _list_summary(self, d):
        r = []
        for e in d.read(False):
            if e.is_dir:
                r.append({
                    'subdir': e.path,
                    'context': None,
                    'child_nodes': self._list_summary(e)
                    })
            else:
                if not e.path.endswith('.beh'): continue
                r.append({
                    'context': self._summary(e),
                    'child_nodes': []
                })
        return r
    
    def _summary(self, t):
        runner = FitRunner(t, self.interpreter)
        return {
            'path': t.path.replace('.beh', '.fit'),
            'title': runner.title,
            'app': self._get_app_name()
        }

class List(Builtin):
    command_decl = u"""
    /**
     * ビヘイビアファイル一覧の出力。
     */
    command list :: void -> void
        reads behaviors
        updates stream
        uses interpreter
        refers python:caty.command.fit.List;
    """
    def execute(self):
        d = self.behaviors.opendir('/')
        r = []
        for e in d.read(True):
            if e.path.endswith('.beh'):
                runner = FitRunner(e, self.interpreter)
                self.stream.write('%s: %s\n' % (runner.path, runner.title))

class ClearReport(Builtin):
    command_decl = u"""
    /**
     * CatyFITの結果の削除。
     * デフォルトではビヘイビアファイルよりも古い結果ファイルをすべて削除する。
     * --allを指定することで、すべての結果ファイルを削除できる。
     *
     * ファイルあるいはディレクトリ名を引数に指定することも可能である。
     * ファイル名を引数に指定した場合、そのファイルに対応した結果ファイルを削除する。
     * ディレクトリ名を指定した場合、そのディレクトリ配下に対して前節の動作を行う。
     */
    command clean {"all": boolean} [string?] :: void -> void
        uses [pub, behaviors]
        reads env
        refers python:caty.command.fit.ClearReport;
    """

    def setup(self, opts, dir_or_file=None):
        self.__all = opts.all
        self.__path = dir_or_file

    def execute(self):
        app = self.env.get('CATY_APP')['name']
        if self.__path:
            path = 'fit-view:/%s/%s' % (app, self.__path.replace('.beh', '.fit'))
            if self.pub.is_dir(self.__path):
                d = self.pub.opendir(path)
                if self.__all:
                    d.delete(True)
                else:
                    self._delete_garbage()
            else:
                f = self.pub.open(path)
                f.delete()
        else:
            d = self.pub.opendir('fit-view:/%s' % app)
            if self.__all:
                d.delete(True)
            else:
                self._delete_garbage()

    def _delete_garbage(self):
        app = self.env.get('CATY_APP')['name']
        if self.__path:
            beh_dir = self.behaviors.opendir(self.__path)
            fit_dir = self.pub.opendir('fit-view:/%s%s' % (app, self.__path))
        else:
            beh_dir = self.behaviors.opendir('/')
            fit_dir = self.pub.opendir('fit-view:/%s' % app)
        for e in beh_dir.read(True):
            if e.path.endswith('.beh'):
                b = self.behaviors.open(e.path)
                f = self.pub.open('fit-view:/%s%s' % (app, e.path.replace('.beh', '.fit')))
                if f.exists and f.last_modified < b.last_modified:
                    f.delete()
        for e in fit_dir.read(True):
            if e.path.endswith('.fit'):
                b = self.behaviors.open(e.path.replace('/'+app, '').replace('.fit', '.beh'))
                f = self.pub.open('fit-view:/%s%s' % (app, e.path.replace('/'+app, '')))
                if not b.exists:
                    f.delete()

