#coding: utf-8
from caty.core.command import Builtin
from caty.core.exception import *
from caty.command.file import FileUtilMixin
import caty.jsontools as json
import caty.jsontools.xjson as xjson
import caty.jsontools.stdjson as stdjson

name = 'json'
schema = u"""
type AnyObject = {*:any};
"""

class ReadJson(FileUtilMixin, Builtin):
    command_decl = u"""
        /**
         * 引数で指定されたファイルを JSON データとして読み込み、値を返す。
         * JSON でないデータの書き込まれたファイルに対しての動作は保証しない（通常はエラー）。
         */
        command read [string] :: void -> any
            reads [pub, data]
            refers python:caty.command.jsonlib.ReadJson;
    """
    def execute(self):
        f = self.open('rb')
        if not f.exists:
            raise FileNotFound(f.path)
        j = stdjson.load(f)
        f.close()
        return j

class WriteJson(FileUtilMixin, Builtin):
    command_decl = u"""
        /**
         * 引数で指定されたファイルに JSON データを書き込む。
         * このコマンドで書き込んだデータは json:read で読み出し可能である。
         */
        command write [string] :: any -> void
            updates [data, pub]
            refers python:caty.command.jsonlib.WriteJson;
    """

    def execute(self, input):
        f = self.open('wb')
        stdjson.dump(input, f)
        f.close()

class Pretty(Builtin):
    command_decl = u"""
        /**
         * 入力値の JSON オブジェクトを整形して返す。
         */
        command pretty :: any -> string
            refers python:caty.command.jsonlib.Pretty;
    """
    def execute(self, input):
        return json.pp(input)

from caty.util.collection import merge_dict
class Merge(Builtin):
    command_decl = u"""
    /**
     * 二つの object をマージする。
     * --mode オプションではプロパティ名が衝突した場合の解消方法を指定する。
     *
     * * "fst": 第一要素のものを用いる
     * * "snd": 第二要素のものを用いる
     * * "error": エラーとして扱い、 null を返す
     */
    command merge {"mode": string} :: [AnyObject, AnyObject] -> AnyObject | null
        refers python:caty.command.jsonlib.Merge;
    """

    def setup(self, opts):
        if not opts.mode:
            self.merge_mode = 'pre'
        else:
            if opts.mode == 'fst':
                self.merge_mode = 'pre'
            elif opts.mode == 'snd':
                self.merge_mode = 'post'
            else:
                self.merge_mode = 'error'

    def execute(self, input):
        try:
            return merge_dict(input[0], input[1], self.merge_mode)
        except:
            return None

class JsonResponse(Builtin):
    command_decl = u"""
    /**
     * 入力値を application/json フォーマットで出力する。
     */
    command response {"status": integer, "encoding": string} :: T -> Response
        refers python:caty.command.jsonlib.JsonResponse;
    """
    def setup(self, opts):
        self._status = opts.status or 200
        self._encoding = opts.encoding or 'utf-8'

    def execute(self, input):
        out = json.dumps(input, ensure_ascii=False)
        return {
            'status': self._status,
            'body': out,
            'header': {
                'content-type': u'application/json; charset=%s' % self._encoding,
            }
        }


class Parse(Builtin):
    command_decl = u"""
    /**
     * 入力値文字列をパースして、JSONオブジェクトを出力する。
     */
    command parse :: string -> any
        reads env
        refers python:caty.command.jsonlib.Parse;
    """
    def execute(self, input):
        return json.loads(input)
