#coding: utf-8
from caty.core.command import Builtin
from caty.jsontools import *
from caty.jsontools.path import build_query

name = 'strg'
schema = u"""
type DumpInfo = {
    "tableName": string,
    "schema": string,
    "data": [object*]
};

type TableSchema = {
    "tableName": string,
    "schema": string,
    "appName": string | null,
};
"""

class StorageAccessor(object):
    u"""JSON ストレージ操作コマンド用の Mix-in クラス。
    コードの簡潔さと処理の重複の回避のために、各コマンドに多重継承を用いて Mix-in する。
    """
    @property
    def table(self):
        if ':' in self._table_name:
            app, table = self._table_name.split(':', 1)
            return self.storage(table, app)
        else:
            return self.storage(self._table_name)


class CreateTable(Builtin):
    command_decl = u"""
    /**
     * JSON スキーマに対応したテーブルの作成。
     * 第一引数はスキーマ名であり、省略不可能である。またこのスキーマは object 型でなければならない。
     * 第二引数はテーブル名であり、省略した場合スキーマ名がテーブル名となる。
     * 既にテーブルが存在する場合何もせず終了する。
     */
    command create-table {"as-global": boolean} [string, string?] :: void -> void
        reads schema
        updates storage
        refers python:caty.command.strg.CreateTable;
    """
    def setup(self, opts, schema_name, table_name=None):
        self._schema_name = schema_name
        self.schema[schema_name]
        self._table_name = table_name if table_name else schema_name
        self._global = opts.as_global

    def execute(self):
        self.storage(self._table_name).create(self._schema_name, self._global)

class DropTable(Builtin, StorageAccessor):
    command_decl = u"""
    /**
     * テーブルの削除。
     */
    command drop-table [string] :: void -> void
        updates storage
        refers python:caty.command.strg.DropTable;
    """
    def setup(self, table_name):
        self._table_name = table_name

    def execute(self):
        self.table.drop()

def _attr_getter(k):
    def _(x):
        v = x
        for n in k.split('.'):
            v = x[n]
        return v

def _cmp_dict(a, b, f):
    return cmp(f(a), f(b))

class ListTable(Builtin):
    command_decl = u"""
    /**
     * テーブル名とそのスキーマ一覧の表示。
     */
    command list-tables :: void -> list<TableSchema>
        reads storage
        refers python:caty.command.strg.ListTable;
    """
    def execute(self):
        r = []
        for x in self.storage.tables:
            r.append({
                'tableName': x['table_name'],
                'schema': x['schema'],
                'appName': x['app'],
                })
        return r

class Select(Builtin, StorageAccessor):
    command_decl = u"""
    /**
     * JSON ストレージの検索。
     * 引数で指定されたテーブルに対して入力値をクエリとして検索を行い、その結果を返す。
     * テーブル以降の引数はオプションであり、検索結果のオブジェクトに含める要素のパスを指定する。
     *
     * --limit オプションは検索結果の最大件数を指定する。デフォルトでは無制限に値を取得する。
     * --offset オプションは検索の開始位置を指定する。デフォルトでは最初から値を取得する。
     * --order-by オプションはソートに用いるプロパティ名を指定する。デフォルトでは insert された順に値を取得する。
     * 
     * 検索クエリには JSON オブジェクトを用いる。
     *
     * {{{
     * {"birth": 1984} | strg:select person 
     * }}}
     *
     * 上記のクエリの結果は、 person テーブルの birth の値が 1984 であるオブジェクトの配列となる。
     * 特に指定されなかった項目に付いては、その値が何であれ取得される。
     *
     * また、以下のように特殊なタグを使って検索する事も出来る。
     *
     * {{{
     * {"name": @_OR [@_LIKE "山田%", @_LIKE 鈴木%"]} | strg:select person
     * }}}
     *
     * この場合は name の値が "山田" か "鈴木" で始まるオブジェクトのリストが検索結果となる。
     *
     * クエリが null の場合、全件取得が行われる。
     *
     * これら特殊タグの一覧を以下に示す。
     *
     * * @_ANY              検索条件なし
     * * @_OR [条件リスト]  [条件リスト]のうちいずれかを満たす
     * * @_AND [条件リスト] [条件リスト]をすべて満たす
     * * @_EACH <条件>      配列の要素全てが条件を満たす（配列型の値に対してのみ使用可能）
     * * @_CONTAINS <条件>  配列の要素が一つでも条件を満たす（配列型の値に対してのみ使用可能）
     * * @_LIKE <パターン>  <パターン>に一致する（string型に対してのみ使用可能）
     * * @_NOT_NULL         null でない 
     * * @_LT <数値>        オブジェクトの値が<数値>より小さい
     * * @_LE <数値>        オブジェクトの値が<数値>より小さいあるいは等しい
     * * @_GT <数値>        オブジェクトの値が<数値>より大きい
     * * @_GE <数値>        オブジェクトの値が<数値>より大きいあるいは等しい
     * 
     */
    command select {
                    "limit": pinteger, 
                    "reverse":boolean, 
                    "order-by": string, 
                    "offset": pinteger,
                   } [string, string*] :: any -> list<_T>
        reads storage
        refers python:caty.command.strg.Select;
    """
    def setup(self, opts, table_name, *args):
        self._table_name = table_name
        self._limit = -1 if not opts.limit else opts.limit
        self._offset = 0 if not opts.offset else opts.offset
        self._order_by = opts.order_by
        self._reverse = opts.reverse
        self._path = args
        l = self.schema['list']
        l.schemata = [self.table.schema]
        self._out_schema = l

    def execute(self, input):
        if not input: input = TagOnly('_ANY')
        r = list(self.table.select(input, self._limit, self._offset, self._reverse))
        if self._order_by:
            r.sort(cmp=lambda a, b:_cmp_dict(a, b, _attr_getter(self._order_by)))
        path = self._path
        if not path:
            return r
        else:
            if len(path) == 1:
                q = build_query('$.'+path[0])
                return [q.find(v).next() for v in r]
            else:
                queries = map(build_query, map(lambda a:'$.' + a, path))
                return [[q.find(v).next() for q in queries] for v in r]


from caty.core.schema.base import TagSchema, NullSchema
class Select1(Builtin, StorageAccessor):
    command_decl = u"""
    /**
     * 一件のデータ取得。
     * クエリの指定の仕方は select と同様である。
     * 検索結果が一件でない場合、 @Error が出力値となる。
     *
     */
    command select1 [string] :: any -> _T | @Error string
        reads storage
        refers python:caty.command.strg.Select1;
    """
    def setup(self, table_name):
        self._table_name = table_name
        self._out_schema = self.table.schema | TagSchema('NG', NullSchema())

    def execute(self, input):
        if not input: input = TagOnly('_ANY')
        r = list(self.table.select(input, -1, 0))
        if len(r) != 1:
            return tagged('Error', u'要素数が1つではありません: %d' % (len(r)))
        return r[0]

class Insert(Builtin, StorageAccessor):
    command_decl = u"""
    /**
     * JSON テーブルに入力値を挿入する。
     * --alow-dup オプションが指定された場合、重複した値を挿入可能になる。
     * デフォルトでは重複した値を挿入しようとした場合は値が挿入されず、 @NG が出力値となる。
     * 
     */
    command insert {"allow-dup": boolean} [string] :: _T -> @OK null | @NG null
        updates storage
        refers python:caty.command.strg.Insert;
    """
    def setup(self, opts, table_name):
        self._table_name = table_name
        self._allow_dup = opts.allow_dup
        self._in_schema = self.table.schema

    def execute(self, input):
        if not self._allow_dup:
            l = list(self.table.select(input, -1, 0))
            if len(l) !=0:
                return tagged('NG', None)
        self.table.insert(input)
        return tagged('OK', None)

class Update(Builtin, StorageAccessor):
    command_decl = u"""
    /**
     * JSON ストレージの更新。
     * 入力値の最初の要素で JSON ストレージを検索し、その結果を第二の入力値で更新する。
     */
    command update [string] :: [any, any] -> null
        updates storage
        refers python:caty.command.strg.Update;
    """
    def setup(self, table_name):
        self._table_name = table_name

    def execute(self, input):
        self.table.update(*input)

class Delete(Builtin, StorageAccessor):
    command_decl = u"""
    /**
     * JSON テーブルの値の削除。
     * 
     */
    command delete [string] :: any -> null
        updates storage
        refers python:caty.command.strg.Delete;
    """
    def setup(self, table_name):
        self._table_name = table_name

    def execute(self, input):
        self.table.delete(input)

class TableDump(Builtin, StorageAccessor):
    command_decl = u"""
    /**
     * テーブルのダンプ。
     * 出力結果は {"tableName":テーブル名, "schema": スキーマ名, "data": データの配列} という形式になる。
     */
    command dump 
        [string] :: void -> DumpInfo
        reads storage
        refers python:caty.command.strg.TableDump;
    """

    def setup(self, table_name):
        self._table_name = table_name

    def execute(self):
        for t in self.storage.tables:
            if t['table_name'] == self._table_name:
                storage = self.storage(t['table_name'])
                return {'tableName': unicode(t['table_name']),
                        'schema': unicode(t['schema']),
                        'data':list(storage.dump())}

class Restore(Builtin):
    command_decl = u"""
    /**
     * ダンプされた情報のリストア。
     * テーブルを新たに生成し、すべてのデータを再度挿入する。
     * 既存のテーブルが存在する場合、一旦ドロップされる。
     */
    command restore :: DumpInfo -> void
        uses storage
        refers python:caty.command.strg.Restore;
    """
    def execute(self, input):
        storage = self.storage(input['tableName'])
        storage.drop()
        storage.create(input['schema'])
        storage = self.storage(input['tableName'])
        storage.restore(input['data'])

