#coding: utf-8
import caty
from caty.util.path import join
from caty import mafs, storage, session
from caty.jsontools import xjson
from caty.util.collection import merge_dict
from pbc import PbcObject, Contract
from caty.command import builtin, debug, file, jsonlib, listlib, secure, strg, test, text, user, fit, xjsonlib
from caty.core.script import node
from caty.core import script, casm
from caty.core.facility import Facility, FakeFacility, FacilitySet, AccessManager
from caty.command.secure import RequestToken, CATY_USER_INFO_KEY
from caty.command.user import User
from caty.util.collection import ImmutableDict
from caty.util import cout
from caty.vcs import BaseClient
from caty.core.async import AsyncQueue

from copy import deepcopy
from operator import truth as _
import locale
import codecs
import operator
import os
import sys

USER = 'main'
ROOT = 'root'
RC_SCRIPT = 'rc.caty'

class System(PbcObject):
    u"""Caty のコアシステム。
    システム起動時に初期化され、内部に全ての Caty アプリケーションを包含する。
    System オブジェクトは一つの Caty プロセスについて一つだけ作成されるものとする。
    """
    __properties__ = ['app_names']

    def __init__(self, encoding=None, is_debug=True):
        cout.write(u'システム情報を読み込んでいます')
        if not encoding:
            encoding = locale.getpreferredencoding()
        self._global_config = GlobalConfig(xjson.load(open('_global.xjson')), 
                                           self._validate_encoding(encoding))
        self._casm = casm.initialize(self,
                                     [builtin, node],
                                     [debug, file, jsonlib, listlib, secure, strg, test, text, user, fit, xjsonlib])
        # アプリケーショングループの順序は rc.caty の実行順序に関わる
        # common, main, extra, examples という順序は固定
        self._app_groups = [
            ag for ag in [
                ApplicationGroup('common', self._global_config, self),
                ApplicationGroup(USER, self._global_config, self),
                ApplicationGroup('extra', self._global_config, self),
                ApplicationGroup('examples', self._global_config, self),
                ApplicationGroup('develop', self._global_config, self),
            ] if ag.exists
        ]
        self._apps = reduce(operator.add, map(ApplicationGroup.apps.fget, self._app_groups))
        self._app_map = {}
        self._debug = is_debug
        if is_debug:
            caty.DEBUG = True
        for app in self._apps:
            if app.name in self._app_map:
                raise Exception(u'アプリケーション名が衝突しています: %s' % (app.name))
            self._app_map[app.name] = app
        cout.write(u'アプリケーションの設定を適用しています')
        for app in self._app_map.values():
            app.finish_setup()
        for group in self._app_groups:
            group.exec_rc_script()
        PbcObject.__init__(self)
        self.__app_names = [app.name for app in self._apps]

    def get_app(self, app_name):
        app = self._app_map[app_name]
        return app

    @property
    def casm(self):
        return self._casm

    @property
    def app_names(self):
        return self.__app_names

    @property
    def server_module_name(self):
        return self._global_config.server_module_name

    @property
    def sysencoding(self):
        return self._global_config.sysencoding

    @property
    def dummy_app(self):
        return DummyApplication(self)

    @property
    def addrsDenied(self):
        return self._global_config._addrsDenied

    @property
    def addrsAllowed(self):
        return self._global_config._addrsAllowed


    @property
    def enable_script_cache(self):
        return self._global_config.enableScriptCache
    
    @property
    def enable_http_method_tunneling(self):
        return self._global_config.enableHTTPMethodTunneling

    def _validate_encoding(self, encoding):
        try:
            codecs.lookup(encoding)
            return encoding
        except LookupError:
            debug.write('Waring: Unknown encdoing %s. Set ascii to system encoding' % self.encdoing)
            return 'ascii'

    def __invariant__(self):
        assert len(self._app_groups) > 0
        assert len(self._apps) > 0
        assert ROOT in self._app_map

    if caty.DEBUG:
        def _contains_name(self, name):
            assert name in self._app_map

        get_app = Contract(get_app)
        get_app.require += _contains_name


class ApplicationGroup(PbcObject):
    u"""Caty アプリケーションをグループ化するオブジェクト。
    管理上の理由により、 Caty のアプリケーションは以下の四つに分類される。

    * main：ユーザ作成のアプリケーション
    * common：フレームワークの提供する共通アプリケーション
    * extra：第三者の配布するアプリケーションの格納場所
    * examples：アプリケーションのサンプル集。本番環境には不要
    
    ApplicationGroup のインスタンスはこれらのアプリケーショングループに対して一つ作成され、
    自身の配下にあるアプリケーションの初期化を行うものとする。
    """

    __properties__ = ['name', 'apps', 'global_config']

    def __init__(self, group_name, global_config, system):
        self._name = group_name
        self._global_config = global_config
        self._system = system
        self._exists = self._make_super_root(self.name).start().read_mode.opendir('/').exists
        if not self.exists:
            if self._name in (USER, 'extra'):
                cout.write(u'アプリケーショングループ %s が存在しません。自動生成します' % self.name)
                d = self._make_super_root('').start()
                d.dual_mode.opendir('/'+self.name).create()
                d.commit()
                self._exists = True
        if self._name == USER:
            d = self._make_super_root(self.name).start()
            r = d.dual_mode.opendir('/root')
            if not r.exists:
                cout.write(u'ルートアプリケーションが存在しません。自動生成します')
                r.create()
                d.commit()
        if self._exists:
            self._apps = list(self._get_apps())
            PbcObject.__init__(self)

    def _get_apps(self):
        if self.exists:
            app_group_root = self._make_super_root(self.name)
            for d in app_group_root.start().read_mode.opendir('/').read():
                if d.is_dir:
                    yield Application(d.path.strip('/'), self, self._system)

    def _make_super_root(self, root):
        return self._global_config.mafs_initializer(DummyApplication(), self._system, ROOT)('.', root, self._global_config.mime_types)

    def exec_rc_script(self):
        for app in self._apps:
            app.exec_rc_script()

    @property
    def name(self):
        return self._name

    @property
    def apps(self):
        return self._apps

    @property
    def global_config(self):
        return self._global_config

    @property
    def exists(self):
        return self._exists

    @property
    def storage_conf(self):
        return self._global_config.storage_conf

    def __invariant__(self):
        if self.exists:
            assert self._name
            assert self._global_config
        else:
            assert self._name
            assert self._global_config
            assert len(self._apps) == 0

        

class Application(PbcObject):
    u"""Caty アプリケーションオブジェクト。
    Caty アプリケーションは自身に特有なファイルシステム、スキーマ、コマンドなどを格納し、
    外部からの入力に応答して出力を返すオブジェクトである（実際の主たる処理は interpreter に委譲する）。

    Caty アプリケーションは他のアプリケーションの参照を行うため、
    互いへの参照を内部的には持っている。
    ただしそれはあくまでも特別なリソースアクセスのために行う事なので、
    基本的に Caty アプリケーション同士は分離された状態で動作する。
    """
    def __init__(self, name, group, system):
        u"""Caty アプリケーションの初期化はブートストラップで行う。
        最初は単にアプリケーション名を設定し、内部的な状態を初期化するだけである。
        """
        self._name = unicode(name)
        self._system = system
        self._path = unicode(name)
        self._group = group
        self._finished = False
        self._physical_path = join(group.name, name)
        self._app_map = {name: self}
        self._global_config = group.global_config
        self._configure()
        cout.write(u'%s の設定を読み込んでいます' % self._path)
        if not self._disabled:
            self._init_appdirs()
            self._init_mafs()
            self._init_schema()
            self._init_session()
            self._init_storage()
            self._init_vcs()
        else:
            cout.write(u'* %s は起動しないように設定されています' % self._name)
        self.async_queue = AsyncQueue(self)
        PbcObject.__init__(self)

    def reload(self):
        self._system._casm.reload_global()
        self.__init__(self._name, self._group, self._system)
        self.finish_setup()

    def exec_rc_script(self):
        if self._disabled: return
        scripts = self._scripts.start()
        if isinstance(scripts, FakeFacility):
            return
        if not scripts.open('/' + RC_SCRIPT).exists:
            return
        cout.write(u'%s の起動スクリプトを実行しています' % self._path)
        facilities = self.create_facilities()
        interpreter = self.interpreter.file_mode(facilities)
        try:
            interpreter.build(RC_SCRIPT)(None)
        except:
            cout.write(u'%s の起動スクリプト実行中にエラーが発生しました' % self.name)

    def add_app(self, app):
        if app.name in app:
            raise Exception(u'アプリケーション名が衝突しています: %s' % app.name)
        self._app_map[app.name] = app

    def get_app(self, name):
        if name not in self._app_map:
            raise Exception(u'存在しないアプリケーション: %s'  % name)
        return self._app_map[name]

    def _configure(self):
        cfg = self._verify_config(self._read_config())
        self._disabled = cfg['disabled']
        self._description = cfg['description']
        self._implLang = cfg['implLang']
        self._assgin = cfg['assign']
        self._indexfiles = cfg.get('indexFiles', ['index.html', 'index.htm', 'index.cgi', 'index.act'])
        self._filetypes = cfg.get('filetypes', {})
        self._encoding = cfg.get('encoding', 'utf-8')
        self._mime_types = self._global_config.mime_types
        self._associations = self._global_config.associations
        self._site_path = self._name
        self._hosturl = self._global_config.host_url
        self._server_module_name = self._global_config.server_module_name
        self._app_spec = AppConfig()
        self._app_spec.update(cfg.get('appProp', {}))
        self._manifest = cfg

    def _read_config(self):
        app_dir = self._group._make_super_root(join(self._group.name, self.name)).start()
        f = app_dir.read_mode.open('/_manifest.xjson')
        if f.exists:
            return xjson.loads(f.read())
        else:
            return self.default_conf()
    
    def default_conf(self):
        return {
            'disabled': False,
            'description': u'%s' % self.name,
            'implLang' : u"python",
            'assign': {
                'pub': u'pub',
                'include': u'include',
                'commands': u'commands',
                'scripts': u'scripts',
                'schemata': u'schemata',
                'behaviors': u'behaviors',
                'data': u'data',
            },
        }

    def _verify_config(self, obj):
        return merge_dict(obj, self.default_conf(), 'pre')

    def _init_appdirs(self):
        app_mafs = self._group._make_super_root(join(self._group.name, self._path)).start()
        app_dir = app_mafs.dual_mode
        generated = []
        for k in self._assgin.keys():
            d = app_dir.opendir('/' + k)
            if not d.exists:
                cout.write(u'%s に %s が存在しません。自動生成します' % (self._path, k))
                d.create()
                generated.append(k)
        app_mafs.commit()
        #if 'pub' in generated:
        #    print u'%s に index.html が存在しません。自動生成します' % (self._path)
        #    pub_mafs = self._group._make_super_root(join(self._group.name, self._path, 'pub')).start()
        #    pub_dir = pub_mafs.dual_mode
        #    index = pub_dir.open('/index.html', 'wb')
        #    index.write(INDEX_HTML)
        #    index.close()
        #    pub_mafs.commit()

    def _init_mafs(self):
        for k, v in self._filetypes.items():
            assoc = v.get('assoc', None)
            if assoc:
                self._associations[k] = assoc
            it = v.get('isText', True)
            ct = v.get('contentType', u'text/plain' if it else u'application/octet-stream')
            ds = v.get('description', u'')
            t = {
                'isText': it,
                'contentType': ct,
                'description': ds,
            }
            self._mime_types[k] = t
        self._associations = ImmutableDict(self._associations)
        root = join(self._group.name, self._path)
        assign = self._assgin
        mafs_init = lambda type, path: self._global_config.mafs_initializer(self, self._system, type)(root, path, self._mime_types, self._encoding)
        self._pub = mafs_init('pub', assign['pub'])
        self._include = mafs_init('include', assign['include'])
        self._data = mafs_init('data', assign['data'])
        self._behaviors = mafs_init('behaviors', assign['behaviors'])
        self._command_fs = mafs_init('commands', assign['commands']).start().read_mode
        self._schema_fs = mafs_init('schemata', assign['schemata']).start().read_mode
        
        # スクリプトファイル
        _scripts = mafs_init('scripts', assign['scripts'])
        if _scripts:
            self._scripts = _scripts
        else:
            self._scripts = FakeFacility()


    def _init_schema(self):
        self._schema_module = self._system.casm.make_app_module(self)
        #self._schema_module.compile()

    def _init_interpreter(self):
        self._interpreter = script.initialize(self._schema_module, self, self._system)
        self._command_finder = self._interpreter.finder

    def _init_session(self):
        self._session_storage = self._global_config.session_storage
    
    def _init_storage(self):
        self._storage = storage.initialize(self._global_config.storage_conf)

    def _init_vcs(self):
        if self._global_config.vcs_module:
            self._vcs = self._global_config.vcs_module.VCSClient(self.pub.start(), self.data.start())
        else:
            self._vcs = BaseClient(self.pub.start(), self.data.start())

    def finish_setup(self):
        if self._disabled: return
        self._schema_module.resolve()
        self._init_interpreter()
        c = {}
        c['indexfiles'] = self.indexfiles[:]
        c['script_ext'] = ['.cgi', '.act', '.do']
        c['associations'] = {}
        for k, v in self.associations.items():
            c['associations'][k] = v
        self._web_config = ImmutableDict(c)
        if self._app_spec:
            if not '$typename' in self._app_spec:
                print u'%s のアプリケーション固有設定の型情報が登録されていません。' % self.name
                raise Exception()
            tn = self._app_spec['$typename']
            cb = self._app_spec.get('$callback', None)
            tp = self._schema_module.schema_finder.start()[tn]
            try:
                tp.validate(self._app_spec)
                self.__exec_callback(cb)
            except Exception, e:
                print u'%s のアプリケーション固有設定の検証に失敗しました。' % self.name
                cout.write(e)
                raise Exception()

    def __exec_callback(self, callback_class_name):
        import copy
        if not callback_class_name: return
        gd = {}
        modname, clsname = callback_class_name.rsplit('.', 1)
        exec 'from %s import %s as _callback' % (modname, clsname) in gd
        cb = gd['_callback'](copy.deepcopy(self._app_spec))

    def __invariant__(self):
        assert self._name == '' or _(self._name)
        assert _(self._group_name)
        assert _(self._global_config)
        assert _(self._physical_path)
        assert _(self._app_map)
        if self._finished:
            pass
        else:
            pass

    def create_facilities(self, session_maker=None):
        from caty.session.value import create_variable
        from caty.env import Env
        env = Env().dual_mode
        var = create_variable().dual_mode
        stream = StreamWrapper(StdStream(self.sysencoding))
        finder = self._schema_module.schema_finder.start()
        facilities = {
            'pub': self._pub.start(),
            'env': env.start(),
            'stream': stream.start(),
            'var': var.start(),
            'session': session_maker().start() if session_maker else self._session_storage.create().start(),
            'scripts': self._scripts.start(),
            'include': self._include.start(),
            'schema': finder,
            'data': self._data.start(),
            'storage': self._storage.set_finder(finder).start(),
            'behaviors': self._behaviors.start(),
            'vcs': self._vcs.start(),
            'schemata': self._schema_fs,
            'config': self._app_spec,
        }
        facilities['token'] = RequestToken(facilities['session'])
        if facilities['session'].exists(CATY_USER_INFO_KEY):
            facilities['user'] = User(facilities['session'].get(CATY_USER_INFO_KEY))
        else:
            facilities['user'] = User({})
        fset = FacilitySet(facilities, self)
        facilities['interpreter'] = self._interpreter.file_mode(fset)
        return fset

    def init_env(self, facilities, is_debug, modes, system, environ={}):
        env = facilities['env']
        env.put('APP_ENCODING', unicode(self.encoding))
        env.put('SYSTEM_ENCODING', unicode(self.sysencoding))
        env.put('APP_PATH', unicode(self.web_path))
        env.put('CATY_APP', {'description':self.description, 'name':self.name, 'path': unicode(self.web_path)})
        env.put('DEBUG', is_debug)
        siteInfoList= []
        for n in system.app_names:
            s = system.get_app(n)
            siteInfoList.append({'description': s.description, 'name': s.name, 'path': unicode(s.web_path)})
        env.put('CATY_APPS', siteInfoList)
        env.put('HOST_URL', unicode(self.host_url))
        env.put('CATY_VERSION', unicode(caty.__version__))
        env.put('CATY_EXEC_MODE', modes)
        env.put('CATY_HOME', unicode(os.getcwd()))
        env.put('APP_MANIFEST', self._manifest)
        env.put('FILE_ASSOC', self.associations)
        if 'CONTENT_TYPE' in environ:
            env.put('CONTENT_TYPE', unicode(environ['CONTENT_TYPE']))
        else:
            env.put('CONTENT_TYPE', u'text/plain')

    @property
    def name(self):
        return self._name

    @property
    def encoding(self):
        return self._encoding

    @property
    def sysencoding(self):
        return self._global_config.sysencoding

    @property
    def web_path(self):
        return u'%s' % self._name if self._name != ROOT else u''

    @property
    def description(self):
        return self._description

    @property
    def host_url(self):
        return self._hosturl

    @property
    def associations(self):
        return self._associations

    @property
    def interpreter(self):
        return self._interpreter

    @property
    def indexfiles(self):
        return self._indexfiles
    
    @property
    def session_storage(self):
        return self._session_storage

    @property
    def pub(self):
        return self._pub

    @property
    def include(self):
        return self._include

    @property
    def data(self):
        return self._data

    @property
    def behaviors(self):
        return self._behaviors

    @property
    def scripts(self):
        return self._scripts

    @property
    def command_finder(self):
        return self._command_finder

    @property
    def schema_finder(self):
        return self._schema_module.schema_finder

    @property
    def web_config(self):
        u"""Web からのアクセスに必要な設定項目
        """
        return self._web_config

    def __invariant__(self):
        assert '/' not in self.name

class GlobalConfig(object):
    u"""Caty システムのサイト固有でない情報。
    mafs バックエンド、JSON ストレージ、セッションなどを指定する。
    
    "mafs": string,
    "session": {
        *: any
    }
    """
    def __init__(self, obj, encoding):
        fs = obj['mafsModule']
        self._mafs_module = fs
        session_conf = obj.get('session', {'type':'memory', 'expire': 3600})
        secret_key = obj['secretKey']
        self._session_storage = session.initialize({'session': session_conf, 'key': secret_key})
        self._filetypes = mafs.metadata.default_types
        self._encoding = obj.get('encoding', u'utf-8')
        self._hosturl = obj['hostUrl']
        self._server_module_name = obj['serverModule']
        self._storage_conf = obj.get('storage', {'type': 'sqlite', 'path': './storage.db'})
        self._sysencoding = unicode(encoding or locale.getpreferredencoding())
        self._mime_types = {}
        self._associations = {}
        self._addrsAllowed = obj.get('addrsAllowed', None)
        self._addrsDenied = obj.get('addrsDenied', [])
        for k, v in self._filetypes.items():
            assoc = v.get('assoc', None)
            if assoc:
                self._associations[k] = assoc
            it = v.get('isText', True)
            ct = v.get('contentType', u'text/plain' if it else u'application/octet-stream')
            ds = v.get('description', u'')
            t = {
                'isText': it,
                'contentType': ct,
                'description': ds,
            }
            self._mime_types[k] = t
        if 'vcsModule' in obj:
            exec 'import %s as caty_vcs_module' % obj['vcsModule']
            self._vcs_module = caty_vcs_module
        else:
            self._vcs_module = None

        self._enableScriptCache = obj.get('enableScriptCache', False)
        self._enableHTTPMethodTunneling = obj.get('enableHTTPMethodTunneling', False)

    def mafs_initializer(self, app, system, type):
        initializer = MafsInitializer(self._mafs_module, app, system, type)
        return (lambda root, target, mime_types, encoding=self.encoding: 
                          initializer(root, target, mime_types, encoding))


    @property
    def session_storage(self):
        return self._session_storage

    @property
    def host_url(self):
        return self._hosturl

    @property
    def encoding(self):
        return self._encoding

    @property
    def sysencoding(self):
        return self._sysencoding

    @property
    def server_module_name(self):
        return self._server_module_name

    @property
    def storage_conf(self):
        return self._storage_conf

    @property
    def filetypes(self):
        return self._filetypes

    @property
    def mime_types(self):
        return deepcopy(self._mime_types)

    @property
    def associations(self):
        return deepcopy(self._associations)

    @property
    def vcs_module(self):
        return self._vcs_module

    @property
    def addrsDenied(self):
        return self._addrsDenied

    @property
    def addrsAllowed(self):
        return self._addrsAllowed

    @property
    def enableScriptCache(self):
        return self._enableScriptCache

    @property
    def enableHTTPMethodTunneling(self):
        return self._enableHTTPMethodTunneling

class MafsInitializer(object):
    u"""mafs 初期化オブジェクト。
    mafs は数あるファシリティの中でももっとも込み入った初期化手順となっている。
    そのため、 MafsInitializer でラッピングして処理の簡略化を図っている。
    """
    def __init__(self, type, app=None, system=None, res_type=None):
        u"""MafsInitializer は最低一つの引数で初期化される。
        type は必須の引数であり、 mafs 実装モジュールの名称を文字列で渡す。
        app は mafs に紐付けられるアプリケーション、 system はそのアプリケーションの属するシステムである。
        res_type は pub, include, schemata など、アプリケーションディレクトリ直下のディレクトリ名か、
        特殊な用途に用いる root のいずれかである。
        """

        exec 'import %s as mafs_impl' % type
        self.mafs_impl = mafs_impl
        self.app = app
        self.system = system
        self.res_type = res_type

    def init(self, root, target, mime_types, encoding):
        u"""初期化処理。新たな初期化済み mafs ファシリティを返す。

        root は初期化対象の mafs の論理的なルートディレクトリであり、 target はその直下の任意のディレクトリである。
        通常のファイルシステムをバックエンドに用いた場合、これは chroot のように働く。
        それに対して RDBMS 上に mafs を構築した場合などは、バックエンドの実装によっては動作が異なる。
        ただし、例えば /main/app/pub というディレクトリツリーが mafs 上にある場合、
        / を root にし、また main を target にした場合は /main をルートとしてファイルアクセスが可能にならなければならない。

        mime_types は拡張子ごとの MIME タイプ情報である。

        encoding はエンコーディング情報である。
        """
        if target is None:return
        return mafs.initialize(self.app,
                               self.system,
                               self.res_type,
                               fs=self.mafs_impl, 
                               root_dir=join(root, target), 
                               mime_types=mime_types, 
                               encoding=encoding)

    def __call__(self, *args, **kwds):
        return self.init(*args, **kwds)

    @property
    def name(self):
        return self._group_name

    @property
    def apps(self):
        return self.apps

    @property
    def global_config(self):
        return self._global_config

class StdStream(object):
    def __init__(self, encoding):
        self.encoding = encoding
        self.out = codecs.getwriter(self.encoding)(sys.stdout)
        self.input = codecs.getwriter(self.encoding)(sys.stdin)

    def write(self, content):
        self.out.write(content)

    def read(self):
        return xjson.loads(self.input.read())

class StreamWrapper(Facility):
    def __init__(self, stream):
        self.__stream = stream
        Facility.__init__(self)
    
    am = AccessManager()

    @am.update
    def write(self, content):
        self.__stream.write(content)

    @am.read
    def read(self):
        return self.__stream.read()

    @property
    def encoding(self):
        return self.__stream.encoding

    def clone(self):
        return self

INDEX_HTML = u"""
<?caty-meta template="smarty"?>
<html>
    <head>
        <title>Welcome</title>
    </head>
    <body>
        {include file="templates:/global_menu.html"}
        <h1>Welcome</h1>
    <!-- 
    This file is automatically generated by Caty.
    The file contains Japanese characters encoded in UTF-8.
    -->
        <p>Caty のサイトにようこそ。</p>

        <p>(これは、Caty によって自動生成されたファイルです。)</p>
        {include file="templates:/footer.html"}
    </body>
</html>
"""


class DummyApplication(Application):
    _name = u''
    def __init__(self, system=None):
        self._system = system
        self.async_queue = AsyncQueue(self)

class AppConfig(FakeFacility, dict):
    pass


