#coding:utf-8
from caty.core.resource import ResourceFinder
from caty.core.facility import ReadOnlyFacility
from caty.core.casm.language.casmparser import iparse
from caty.core.casm.language.ast import ModuleName
from topdown import *

class SchemaFinder(ResourceFinder, ReadOnlyFacility):
    u"""スキーマ検索オブジェクト。
    通常のコマンドクラスから参照されるので、ファシリティとして扱う。
    """
    def __init__(self, module, app, system):
        self._module = module
        ResourceFinder.__init__(self, app, system)
        self._local = LocalModule(self)

    def add_local(self, string):
        u"""ローカルスキーマの追加は一見すると破壊的操作だが、
        実際には永続的な効果を及ぼさず、またそもそも CatyFIT など限られた用途にしか用いない。
        そのため、書き込み処理としては扱わない。
        """
        self._local.compile(string)

    def __getitem__(self, key):
        key = key.lstrip(':')
        c = key.count(':')
        if  c == 2:
            app_name, scm = key.split(':', 1)
            if app_name == self.application.name:
                return self[scm]
            if app_name in self.system.app_names:
                app = self.system.get_app(app_name)
                return app.schema_finder[scm]
        else:
            if key.startswith(self._local.name+':'):
                k = key.split(':')[1]
                if self._local.has_schema(k):
                    return self._local.get_schema(k)
            elif self._local.name == 'public':
                if self._local.has_schema(key):
                    return self._local.get_schema(key)
            return self._module.get_schema(key)

    def __call__(self, key):
        return self[key]

    def clone(self):
        return self

class LocalModule(ResourceFinder):
    def __init__(self, finder):
        self.schema_ns = {}
        self.name = u''
        self.schema_finder = OverlayedFinder(self, finder)
        ResourceFinder.__init__(self, finder.application, finder.system)
        self._lazy_resolvers = []

    def add_lazy_resolve(self, f):
        self._lazy_resolvers.append(f)

    def has_schema(self, name):
        return name in self.schema_ns 

    def get_schema(self, name):
        if self.has_schema(name):
            return self.schema_ns[name]
        else:
            return self.schema_finder[name]

    def compile(self, schema_string):
        u"""組み込みコマンドの型指定など、オンザフライでスキーマを構築するためのメソッド。
        """
        def modname(cs):
            cs.parse(until('m'))
            cs.parse('module')
            cs.parse(many(' '))
            n = cs.parse(until(';'))
            return n.strip()
        self.name = as_parser(modname).run(schema_string)
        for t in iparse(schema_string):
            t.build_schema_tree(self)

        for s in self.schema_ns.values():
            try:
                s.resolve(self)
            except:
                print 'DEBUG', s.name
                raise
        for l in self._lazy_resolvers:
            l.resolve()
        self._lazy_resolvers = []

    def add_schema(self, name, schema, annotation):
        self.schema_ns[name] = schema

class OverlayedFinder(object):
    
    def __init__(self, module, finder):
        self.module = module
        self.finder = finder

    def __getitem__(self, key):
        if self.module.has_schema(key):
            return self.module.get_schema(key)
        else:
            return self.finder[key]

    __call__ = __getitem__

class CommandFinder(dict, ResourceFinder, ReadOnlyFacility):
    u"""コマンド検索オブジェクト。
    こちらもコマンドから参照される。
    """
    def __init__(self, commands, app, system):
        dict.__init__(self)
        ResourceFinder.__init__(self, app, system)
        dict.update(self, commands)

    def __getitem__(self, key):
        c = key.count(':')
        if c == 2: #別のアプリケーションを参照している場合
            app_name, cmd = key.split(':', 1)
            if app_name in self.system.app_names:
                app = self.system.get_app(app_name)
                return app.command_finder[cmd]
        else:
            return dict.__getitem__(self, key.lstrip(':'))
        raise KeyError(key)

    def __setitem__(self, key, value):
        raise Exception(u'実行時でのコマンド追加は不可能です')

    def __delitem__(self, key):
        raise Exception(u'実行時でのコマンド削除は不可能です')

    def update(self, arg):
        raise Exception(u'実行時でのコマンド追加は不可能です')



