#coding: utf-8
from caty.core.schema import schemata, TypeVariable, NamedSchema, PseudoTag, UnionSchema
from caty.core.command.profile import CommandProfile, ProfileContainer

class Provide(object):
    def __init__(self, names):
        self.exports = names

    def build_schema_tree(self, module):
        module.exports= self.exports

class ModuleName(object):
    def __init__(self, name):
        self.name = name

    def build_schema_tree(self, module):
        if self.name == 'builtin': #ビルトインのみは特別扱い
            if not module.is_builtin:
                raise Exception(u'"builtin" は組み込みのモジュールにのみ適用できます')
        else:
            if self.name != module.name:
                raise Exception(u'モジュール名 "%s" はパス名 "%s" に適用できません' % (self.name, module.filepath))

class LazyResolvedReference(NamedSchema):
    def __init__(self, type_ref, args, kwds):
        self.type_ref = type_ref
        self.args = args
        self.kwds = kwds
        self._name = type_ref._name
        self._type_vars = type_ref._type_vars
        self.__resolved = False

    def apply(self, type_args):
        self.type_args = type_args
        return self

    def resolve(self):
        NamedSchema.__init__(self, *self.type_ref._lazy_resolve(self.module, self.type_args, *self.args, **self.kwds))
        self.__resolved = True
    
    # XXX: 極めて実験的
    def union(self, another):
        if self.__resolved:
            return NamedSchema.union(self, another)
        else:
            return UnionSchema(self, another)

    def __invariant__(self):
        pass

    def clone(self):
        return self

    @property
    def type(self):
        return self.type_ref.type

class TypeReference(NamedSchema):
    def __init__(self, name, type_vars, definition, annotation):
        self._name = name
        self._type_vars = type_vars if type_vars else ()
        self.__definition = definition
        self.__annotation = annotation
        self.__marked = False
        self.__resolved = False

    @property
    def definition(self):
        return self.__definition

    def build_schema_tree(self, module):
        self.module = module
        module.add_schema(self._name, self, self.__annotation)

    def resolve(self, module):
        if not self.__marked:
            self.__marked = True
            schema = self.definition.build(self.module)
            NamedSchema.__init__(self, self._name, self._type_vars, schema, module)
            self.__resolved = True
    
    def _lazy_resolve(self, module, type_args, *args, **kwds):
        schema = self.definition.build(module).clone(*args, **kwds).apply(type_args)
        self.__resolved = True
        return self._name, self._type_vars, schema, module

    @property
    def marked(self):
        return self.__marked

    @property
    def resolved(self):
        return self.__resolved

    @property
    def type(self):
        if not self.resolved:
            return '__type_ref__'
        else:
            return NamedSchema.type.fget(self)

    def clone(self, *args, **kwds):
        if self.resolved:
            return NamedSchema.clone(self, *args, **kwds)
        else:
            return LazyResolvedReference(self, args, kwds)


class Node(object):
    def __init__(self):
        self.options = {}

    def optional():
        def _get(self):
            if 'optional' in self.options:
                return self.options['optional']
            else:
                return False
        def _set(self, value):
            self.options['optional'] = value
        return _get, _set
    optional = property(*optional())

class ScalarNode(Node):
    def __init__(self, schema_name, options=None, type_args=None):
        Node.__init__(self)
        self.schema_name = schema_name
        self.options = options if options is not None else {}
        self.type_args = type_args if type_args is not None else []

    def build(self, module):
        try:
            schema = module.schema_finder(self.schema_name)
        except KeyError:
            return TypeVariable(self.schema_name)
        ta = {}
        for name, t in zip(iter(schema.type_vars), iter(self.type_args)):
            ta[name] = t.build(module)
        if isinstance(schema, TypeReference):
            schema.resolve(module)
        ns = schema.clone(options=self.options).apply(ta)
        if isinstance(ns, LazyResolvedReference):
            ns.module = module
            module.add_lazy_resolve(ns)
        return ns

class TypeVarNode(Node):
    def __init__(self, type_name):
        Node.__init__(self)
        self.type_name = type_name

    def build(self, module):
        schema = TypeVariable(self.type_name)
        return schema

class OperatorNode(Node):
    def __init__(self, node1, node2):
        Node.__init__(self)
        self.node1 = node1
        self.node2 = node2

    def build(self, module):
        s1 = self.node1.build(module)
        s2 = self.node2.build(module)
        s = self.operate(s1, s2)
        if 'optional' in self.options:
            s.optional = self.options['optional'] 
        return s
    
import operator
class UnionNode(OperatorNode):
    operate = operator.or_

class IntersectionNode(OperatorNode):
    operate = operator.and_

class UpdatorNode(OperatorNode):
    operate = operator.lshift

class TaggedNode(Node):
    def __init__(self, tag, node):
        self.tag = tag
        self.schema_node = node if node else ScalarNode('any')

    def build(self, module):
        schema = self.schema_node.build(module)
        return schema.to_tagged(self.tag)

    def optional():
        def _get(self):
            return 'optional' in self.schema_node.options and self.schema_node.options['optional']
        def _set(self, value):
            self.schema_node.options['optional'] = value
        return _get, _set
    optional = property(*optional())

class PseudoTaggedNode(Node):
    def __init__(self, name, value, node):
        self._name = name
        self._value = value
        self.schema_node = node

    def build(self, module):
        schema = self.schema_node.build(module)
        schema.pseudoTag = PseudoTag(self._name, self._value)
        return schema

    def optional():
        def _get(self):
            return 'optional' in self.schema_node.options and self.schema_node.options['optional']
        def _set(self, value):
            self.schema_node.options['optional'] = value
        return _get, _set
    optional = property(*optional())


class EnumNode(Node):
    def __init__(self, enum):
        Node.__init__(self)
        #t = type(enum[0])
        #if not all(map(lambda x:t==type(x), enum)):
        #    raise ValueError(t)
        self.enum = enum

    def build(self,module):
        schema = module.get_schema('enum')
        return schema.clone(self.enum, {'optional': self.optional})

class ArrayNode(ScalarNode):
    def __init__(self, items, options):
        ScalarNode.__init__(self, 'array', options, [])
        self.items = items
        self.options = options

    def build(self, module):
        schema = module.get_schema(self.schema_name)
        items = []
        for i in self.items:
            items.append(i.build(module))
        return schema.clone(self.options, items)

class ObjectNode(ScalarNode):
    def __init__(self, items=None, wildcard=None, options=None):
        ScalarNode.__init__(self, 'object', {}, [])
        self.items = items if items else {}
        self.wildcard = wildcard if wildcard else ScalarNode('never')
        self.options = options if options else {}

    def build(self, module):
        items = {}
        for k, v in self.items.iteritems():
            items[k] = v.build(module)
        schema = module.get_schema(self.schema_name)
        w = self.wildcard.build(module)
        schema = schema.clone(items, w, self.options)
        return schema

class CommandNode(object):
    def __init__(self, name, patterns, uri, doc, annotation):
        self.name = name
        self.patterns = patterns
        self.uri = [u for l, u in uri if l == 'python:'][0]
        self.doc = doc
        self.annotation = annotation
        self.profile_container = None

    def build_schema_tree(self, module):
        module.add_command(self.name, self)
        self.application = module.application

    def resolve(self, module):
        if self.profile_container:
            module.replace_command(self.name, self.profile_container)
            return
        pc = ProfileContainer(self.name, self.uri, module.command_loader, self.annotation, self.doc, self.application)
        for p in self.patterns:
            p.build(module)
            pc.add_profile(CommandProfile(p.opt_schema, p.arg_schema, p.decl))
        module.replace_command(self.name, pc)
        self.profile_container = pc

class CallPattern(object):
    def __init__(self, opts, args, decl):
        self.opts = opts
        self.args = args
        self.decl = decl

    def build(self, module):
        if self.opts:
            self.opt_schema = self.opts.build(module)
        else:
            self.opt_schema = schemata['null']
        if self.args:
            self.arg_schema = self.args.build(module)
        else:
            self.arg_schema = schemata['null']
        self.decl.build(module)

class CommandDecl(object):
    u"""コマンドの宣言部分の構造体。以下の要素それぞれのリストを持つ構造体クラスである。
    型宣言 type -> type
    大域脱出宣言 throws, breaks
    資源使用宣言 uses, reads, updates
    コマンドURI参照 refers
    """
    def __init__(self, profiles, jump, resource):
        self.uri = '' # 後で挿入される
        self.profiles = profiles
        self.jump = jump if isinstance(jump, list) else [jump]
        self.resource = resource if isinstance(resource, list) else [resource]

    def get_all_resources(self):
        for res in self.resource:
            t = res[0]
            for r in res[1]:
                yield t, r

    def build(self, module):
        p = []
        for i, o in self.profiles:
            if hasattr(i, 'build'):
                i = i.build(module)
            if hasattr(o, 'build'):
                o = o.build(module)
            p.append((i, o))
        self.profiles = p

        j = []
        for t, ls in self.jump:
            l = []
            for node in ls:
                if hasattr(node, 'build'):
                    node = node.build(module)
                l.append(node)
            j.append(l)

        self.jump  = j


