#coding:utf-8
from __future__ import with_statement
from caty.core.casm.finder import SchemaFinder
from caty.core.casm.loader import CommandLoader, BuiltinLoader
from caty.core.schema import schemata
from caty.core.casm.language.casmparser import iparse
from caty.core.casm.language import xcasmparser as xcasm
from caty.core.casm.plugin import PluginMap
from caty.jsontools.path.validator import to_decl_style
from threading import RLock
from StringIO import StringIO

class Module(object):
    def __init__(self, app):
        self.schema_ns = {}
        self.command_ns = {}
        self.command_loader = None
        self.sub_modules = {}
        self.parent = None
        self._name = u''
        self._app = app
        self._system = app._system
        self._lazy_resolve = []
        self._plugin = PluginMap()
    
    @property
    def application(self):
        return self._app

    @property
    def system(self):
        return self._system

    @property
    def name(self):
        return self._name
 
    def add_schema(self, name, member, annotation):
        self.schema_ns[name] = member

    def add_command(self, name, profile):
        if name in self.command_ns:
            raise Exception('Already registered: %s' % name)
        self.command_ns[name] = profile

    def replace_command(self, name, profile):
        self.command_ns[name] = profile
 
    def add_lazy_resolve(self, f):
        self._lazy_resolve.append(f)

    def resolve(self):
        u"""型参照の解決
        """
        for s in self.schema_ns.values():
            try:
                s.resolve(self)
            except:
                print 'DEBUG', s.name
                raise
        for c in self.command_ns.values():
            try:
                c.resolve(self)
            except:
                print 'DEBUG', c.name
                raise
        for m in self.sub_modules.values():
            try:
                m.resolve()
            except:
                print 'DEBUG', m.name
                raise
        for l in self._lazy_resolve:
            l.resolve()
        self._lazy_resolve = []

    def has_schema(self, name, tracked=()):
        if self in tracked:
            return False
        t = list(tracked ) + [self]
        if name in self.schema_ns:
            return True
        else:
            if ':' in name:
                m, n = name.rsplit(':', 1)
                if m in self.sub_modules and self.sub_modules[m].has_schema(n, t):
                    return True
            if self.parent:
                return self.parent.has_schema(name, t)
            else:
                return False

    def has_command(self, name, tracked=()):
        if self in tracked:
            return False
        t = list(tracked ) + [self]
        if name in self.command_ns:
            return True
        else:
            if ':' in name:
                m, n = name.rsplit(':', 1)
                if m in self.sub_modules and self.sub_modules[m].has_command(n, t):
                    return True
            if self.parent:
                return self.parent.has_schema(name, t)
            else:
                return False


    def get_schema(self, name):
        if not self.has_schema(name):
            raise KeyError(name)
        if name in self.schema_ns:
            return self.schema_ns[name]
        else:
            if ':' in name:
                m, n = name.rsplit(':', 1)
                if m in self.sub_modules:
                    return self.sub_modules[m].get_schema(n)
        return self.parent.get_schema(name)

    def get_command_type(self, name):
        if not self.has_command(name):
            raise KeyError(name)
        if name in self.command_ns:
            return self.command_ns[name]
        for m in self.sub_modules.values():
            if m.has_command(name):
                return m.get_command(name)
        return self.parent.get_command_type(name)

    @property
    def command_profiles(self):
        for k, v in self.command_ns.iteritems():
            yield k, v
        for k, m in self.sub_modules.iteritems():
            for a, b in m.command_profiles:
                if k == 'public':
                    yield a, b
                else:
                    yield k + ':' + a, b

    @property
    def schema_finder(self):
        return SchemaFinder(self, self._app, self._system)

    def get_plugin(self, name):
        return self._plugin.get_plugin(name)

class CoreModule(Module):
    u"""Caty付属のモジュール。
    コマンドクラスはすべてライブラリのサーチパス上にあり、
    スキーマはそれらの.pyファイルのdocstringなどから取得される。
    """
    def __init__(self, app, parent, is_root=False):
        Module.__init__(self, app)
        self.schema_ns = schemata
        self.command_loader = BuiltinLoader()
        self.parent = parent
        self.is_root= is_root
        self.is_builtin = True
        self.filepath = u''

    def compile(self, schema_string, module):
        if self.is_root:
            if module is None:
                self._name = 'builtin'
                for t in iparse(schema_string):
                    t.build_schema_tree(self)
            else:
                mod = self.__class__(self._app, self)
                mod.compile(schema_string, module)
                self.sub_modules[mod.name] = mod
        else:
            self._name = module.name
            for t in iparse(schema_string):
                t.build_schema_tree(self)

class AppModule(Module):
    def __init__(self, app, parent, core=None, global_module=None, is_root=False):
        Module.__init__(self, app)
        self.fs = app._schema_fs
        self.pcasm_cache = None
        self.parser_cache = None
        self.command_loader = CommandLoader(app._command_fs, global_module.command_loader if global_module else None)
        self.parent = parent
        self.is_root= is_root
        self.is_builtin = False
        self._plugin.set_fs(app._command_fs)
        if core:
            for k, v in core.schema_ns.items():
                self.schema_ns[k] = v
            for k, v in core.sub_modules.items():
                self.sub_modules[k] = v
            for k, v in core.command_ns.items():
                self.command_ns[k] = v
        if global_module:
            for k, v in global_module.schema_ns.items():
                self.schema_ns[k] = v
            for k, v in global_module.sub_modules.items():
                self.sub_modules[k] = v
            for k, v in global_module.command_ns.items():
                self.command_ns[k] = v


    def _path_to_module(self, path):
        p = path[1:].rsplit('.', 1)[0]
        return p.replace('/', '.')

    def _module_to_path(self, modname):
        m = modname.rsplit('.', 1)[0]
        return '/' + m.replace('.', '/')

    def compile(self):
        for e in self.find(self.fs.DirectoryObject('/')):
            if self.is_root and e.path == '/public.casm':
                self.filepath = e.path
                self._name = self._path_to_module(self.filepath)
                self._compile(e.path)
            elif e.path.endswith('.casm') or e.path.endswith('.pcasm'):
                mod = self.__class__(self._app, self)
                mod.filepath = e.path
                mod._name = self._path_to_module(mod.filepath)
                mod._compile(e.path)
                self.sub_modules[mod.name] = mod
            elif e.path == '/formats.xjson':
                o = self.fs.open(e.path)
                self._plugin.feed(o.read())

    def _compile(self, path):
        o = self.fs.open(path)
        if path.endswith('.casm'):
            for t in self.find_or_create_cache(o):
                t.build_schema_tree(self)
        elif path.endswith('.pcasm'):
            d = to_decl_style(o)
            if self.pcasm_cache:
                io = self.pcasm_cache
            else:
                io = StringIO(d)
                io.path = o.path
                self.pcasm_cache = io
            for t in self.find_or_create_cache(io):
                t.build_schema_tree(self)
        elif path.endswith('.xcasm'):
            for t in self.find_or_create_cache(o, 'xcasm'):
                t.build_schema_tree(self)
        else:
            pass

    def find_or_create_cache(self, fo, type='casm'):
        if not self.parser_cache:
            if type == 'casm':
                self.parser_cache = iparse(fo.read())
                self._show_msg(fo)
            elif type == 'xcasm':
                self.parser_cache = xcasm.iparse(fo.read())
        return self.parser_cache

    def find(self, do):
        for e in do.read():
            if not e.is_dir:
                yield e
            else:
                for se in self.find(e):
                    yield se

    def _show_msg(self, fo):
        print u'  * スキーマ: %s' % fo.path.strip('/')

class GlobalModule(AppModule):
    u"""${CATY_HOME}/globla/に存在する、すべてのアプリケーションから参照可能なモジュール。
    """
    def __init__(self, app, parent=None, is_root=False):
        Module.__init__(self, app)
        self.fs = GlobalFs('schemata')
        self.pcasm_cache = None
        self.parser_cache = None
        self.command_loader = CommandLoader(GlobalFs('commands'))
        self.parent = parent
        self.is_root= is_root
        self.is_builtin = False
        for k, v in schemata.items():
            self.schema_ns[k] = v
        print u'グローバルコマンド及びスキーマを読み込んでいます'

    def _show_msg(self, dummy):
        pass

class GlobalFs(object):
    def __init__(self, target):
        self.DirectoryObject = GlobalDir(target)
        self.open = GlobalOpener(target)

import os
from caty.util.path import join
class GlobalDir(object):
    def __init__(self, target):
        self._cwd = os.getcwd()
        self._target = target

    def __call__(self, dummy):
        return self

    def read(self, dummy=True):
        for r, d, f in os.walk(join(self._cwd, 'global', self._target)):
            for e in f:
                yield GlobalFile(r, '/'+e)

class GlobalOpener(object):
    def __init__(self, target):
        self._cwd = os.getcwd()
        self._target = target

    def __call__(self, path):
        return GlobalFile(join(self._cwd, 'global', self._target), path)

class GlobalFile(object):
    def __init__(self, base, path):
        self.base = base
        self.path = path
        self.is_dir = False

    def read(self):
        return open(join(self.base, self.path)).read()

class IntegratedModule(object):
    u"""標準モジュールとユーザ定義モジュールの複合モジュール
    """
    def __init__(self, system):
        self._core = CoreModule(system.dummy_app, None, True)
        self._global = GlobalModule(system.dummy_app, None, True)
        self._system = system
    
    def compile(self, *args):
        self._core.compile(*args)
        self._global.compile()
    
    def make_app_module(self, app):
        app_module = AppModule(app, None, self._core, self._global, True)
        app_module.compile()
        return app_module

    def reload_global(self):
        self._global = GlobalModule(self._system.dummy_app, None, True)
        self._global.compile()

