# coding:utf-8
from __future__ import with_statement
import caty.jsontools as json
from optparse import OptionParser
from itertools import *
from functools import partial
from caty.core.command.exception import *
import types

__all__ = ['Command', 'Builtin', 'ScriptError', 'PipelineInterruption', 'PipelineErrorExit', 'compile_builtin']

def execption_hook(f):
    u"""コマンドでのエラー発生時にそのコマンド名を出力し、
    そのまま例外をスローする。
    """
    def wrapped(self, *args, **kwds):
        try:
            return f(self, *args, **kwds)
        except Exception, e:
            if isinstance(e, PipelineInterruption) or isinstance(e, PipelineErrorExit):
                raise
            print "Error:", self
            raise
    return wrapped

class Command(object):
    u"""コマンド基底クラス。
    すべてのコマンドは入力チャンネルと出力チャンネルを持ち、
    関数のように呼び出し可能であるとする。
    このクラスのサブクラスは profile と profile_container の二つのプロパティを持つ。
    このプロパティの実体は caty.schema.commandprofile で定義された
    コマンド型宣言オブジェクトであり、コマンドの入出力の型を定義する。

    これはシステムの初期化時に /schema/ 以下の .casm ファイルで定義された
    コマンドの型指定からインジェクトされる。
    そのため、コマンドクラスの実装に際しては必ずスキーマを書かなければならない。

    schema, env プロパティはそれぞれスキーマモジュールの各関数/クラスへの参照と
    環境変数へのアクセスオブジェクトであり、これらは常にインジェクトされる。
    他のプロパティは入出力をのぞき、基本的にコマンド型宣言の reads, updates, uses 句に
    基づいてインジェクトされる。型宣言に記述の無い Caty ファシリティにはアクセスできない。
    """

    def __init__(self, opt_list, arg_list):
        opts, args, profile = self.profile_container.determine_profile(opt_list, arg_list)
        annotations = self.profile_container.get_annotations()
        self._profile = profile
        self._annotations = annotations
        self._mode = set([u'console', u'web', u'test']).intersection(annotations)
        self._opts = opts
        self._args = args
        self._in_schema = profile.in_schema.clone()
        self._out_schema = profile.out_schema.clone()
        self._defined_application = self.profile_container.defined_application
        self.async_queue = self._defined_application.async_queue
        self.__facility_names = set()

    @property
    def profile(self):
        u"""コマンドの入出力のスキーマ情報オブジェクト。
        入出力のスキーマは多相型への対応のために不変オブジェクトになっていないが、
        基本的に値の変更は行わないこと。
        """
        return self._profile

    @property
    def annotation(self):
        u"""コマンドに対するアノテーション。
        アノテーションではコマンドがどのような条件で動作するかといった情報を付け加える。
        """
        return self._annotation

    def set_facility(self, facilities):
        u"""ファシリティの設定。
        フレームワーク側で行う処理なので、一般のコマンド実装者が直に使うべきではない。
        """
        _set = set()
        for mode, name in self.profile.facilities:
            facility = facilities[name]
            if mode == 'reads':
                facility = facility.read_mode
            elif mode == 'updates':
                facility = facility.update_mode
            elif mode == 'uses':
                facility = facility.dual_mode
            # Dummy で定義されてない（=ユーザ定義）ファシリティは、
            # 定義元のアプリケーションへの参照に差し替える
            if self._defined_application.name and not 'volatile' in self._annotations:
                facility.application = self._defined_application
            setattr(self, name, facility)
            _set.add(name)
        # 互換性維持のため、envとschemaを追加
        if not 'env' in _set:
            setattr(self, 'env', facilities['env'].read_mode)
            _set.add('env')
        if not 'schema' in _set:
            setattr(self, 'schema', facilities['schema'].read_mode)
            _set.add('schema')
        
        # アプリケーション固有設定はすべてのコマンドから読み取り可能
        if not 'config' in _set:
            setattr(self, 'config', facilities['config'].read_mode)
            _set.add('config')
        self.__facility_names = _set

    def get_facility(self, facility_name):
        u"""ファシリティの取得。
        指定されたファシリティが存在しない場合、 None を返す。
        """
        if hasattr(self, facility_name):
            return getattr(self, facility_name)
        else:
            return None

    def _prepare(self):
        opts = self._opts
        args = self._args
        if opts is not None and args:
            self.setup(opts, *args)
        elif opts:
            self.setup(opts)
        elif args:
            self.setup(*args)
        else:
            self.setup()

    def setup(self): pass

    @property
    def in_schema(self):
        u"""入力スキーマ。
        """
        return self._in_schema

    @property
    def out_schema(self):
        u"""出力スキーマ。
        """
        return self._out_schema

    @execption_hook
    def __call__(self, input):
        if self._mode: # @console など、特定のモードでしか動かしてはいけないコマンドのチェック処理
            mode = self.env.get('CATY_EXEC_MODE')
            if not self._mode.intersection(set(mode)):
                raise Exception(u'%s は動作モード %s では使用できません' % (self.__class__.__name__, str(mode)))
        self.in_schema.validate(input)
        if self.profile.in_schema.type == 'void':
            r = self.execute()
        else:
            r = self.execute(input)
        self.out_schema.validate(r)
        if 'commit-point' in self.profile_container.get_annotations():
            for n in self.__facility_names:
                getattr(self, n).commit()
        return r

    def execute(self, *args, **kwds):
        raise NotImplementedError

    def error(self, json_obj):
        u"""コマンドをエラーとして終了させる。
        このメソッドが呼ばれた後はエラーハンドラーに処理が委譲され、
        後のパイプラインは一切処理されなくなる。
        """
        raise PipelineErrorExit(json_obj)

    def exit(self, json_obj):
        u"""コマンドを json_obj を出力にして終了する。
        """
        raise PipelineInterruption(json_obj)

    def clone(self):
        raise NotImplementedError()

    def _new_facility(self):
        fs = {}
        for n in self.__facility_names:
            f = getattr(self, n)
            fs[n] = f.start()
        return fs

    def _internal_commit(self):
        for n in self.__facility_names:
            f = getattr(self, n)
            f.commit()
                
class Builtin(Command):
    u"""組み込みコマンドのベースクラス。
    組み込みコマンドはコマンド宣言をファイルではなく
    プロパティとして直に記述する。
    この時は文字列としてスキーマを記述し、それをスキーマオブジェクトに変換、
    クラスの属性として追加という流れになる。

    このクラス自体は特に何もせず、マーカーとしてのみ使う。
    """

class Internal(Builtin):
    u"""内部コマンドクラス。
    一部のコマンドは Caty システムに直接触る必要がある。
    そのようなコマンドはこのクラスを継承すること。
    ユーザ定義コマンドは基本的に使ってはいけない。
    """
    def set_facility(self, facilities):
        self._facilities = facilities
        self._system = facilities.system
        Builtin.set_facility(self, facilities)

class Syntax(Builtin):
    u"""Caty スクリプトの構文要素のベースクラス。
    Caty スクリプトではオブジェクトやリストの構築、分岐処理もコマンドとして実装する。
    それらのコマンドは通常とは異なったインスタンス化の経路を必要とするため、
    このクラスで必要な処理を提供する。
    """
    def __init__(self):
        Builtin.__init__(self, None, None)

def compile_builtin(module, registrar):
    u"""ビルトインのコマンドのプロパティで定義されたコマンド宣言をコンパイルする。
    ビルトインのコマンドはすべて以下のプロパティを持っていなければならない。
    
    * command_decl

    このプロパティにはコマンド宣言を直に記述する。
    """
    for item in filter(lambda a: isinstance(a, types.TypeType), module.__dict__.values()):
        if issubclass(item, Builtin) and item != Builtin:
            schema_string = u"%s" % (item.command_decl)
            registrar.compile(schema_string)

