#coding: utf-8
from caty.core.schema.base import NullSchema
from caty.util import debug
from caty.util.optionparser import OptionParser, OptionParseFailed
from itertools import *
from functools import *

class ProfileContainer(object):
    u"""複数の CommandProfile をまとめるクラス。
    オプション引数位置固定引数で CommandProfile を決定する。
    入出力の型がオーバーロードされている場合、
    その後の処理は CommandProfile で行う。
    """
    def __init__(self, name, uri, commands, annotation, doc, app):
        self.profiles = []
        self.name = name
        path = (uri.split(':')[-1])
        pkgname, cmdname = path.rsplit('.', 1)
        self.command_class = commands.get(pkgname, cmdname)
        self.annotation = annotation or u''
        self.doc = doc if doc else 'undocumented'
        self.defined_application = app

    def get_annotations(self):
        return [a.strip() for a in self.annotation.split(',')]

    def add_profile(self, profile):
        self.profiles.append(profile)

    def determine_profile(self, opts, args):
        lasterror = None
        last_tb = None
        import traceback
        for p in self.profiles:
            try:
                o, a = p.parse_option(opts, args)
                return o, a, p
            except Exception, e:
                lasterror = e
                last_tb = traceback.format_exc()
        debug.write(unicode(last_tb, 'utf-8'))
        raise lasterror
    
    def get_command_class(self):
        return self.command_class

    @property
    def type_info(self):
        r = []
        for p in self.profiles:
            opts, args, input, output = p.usage
            if opts:
                r.append('Usage: %s [OPTION] <%s>' % (self.name, args))
                r.append('Option:\n%s' % self.indent(opts))
            else:
                if args == 'null':
                    r.append('Usage: %s' % (self.name))
                else:
                    r.append('Usage: %s <%s>' % (self.name, args))
            r.append('Input:\n%s' % self.indent(input))
            r.append('Output:\n%s' % self.indent(output))
            r.append('\n')
        return '\n'.join(r)

    def resolve(self, module):
        for p in self.profiles:
            p.resolve(module)

    @property
    def usage(self):
        return self.type_info + 'Description:\n' + self.doc

    @property
    def title(self):
        return self.doc.splitlines()[0].strip()

    def indent(self, s):
        r = []
        for l in s.splitlines():
            r.append('    ' + l)
        return '\n'.join(r)

class CommandProfile(object):
    u"""コマンドの引数や型の宣言を定義するクラス。
    一つのコマンドが複数の引数タイプなどでオーバーロードされるため、
    このクラス経由でコマンド確定時の入出力のスキーマを決定する。
    また、以下のようにコマンドの型宣言には複数の入出力のプロファイルがありうる。

        command foo {} [] :: null -> string,
                             string -> string
                             uses some_facility;

    この場合入力が null のものがデフォルトで採用される。
    また、プロファイルが一つしかない場合は常にそのプロファイルを利用する。
    """
    def __init__(self, opts_schema, args_schema, declobj):
        self.opts_schema = opts_schema
        self.args_schema = args_schema
        self.declobj = declobj
        self._in_schema = None
        self._out_schema = None
        if len(declobj.profiles) == 1:
            self._in_schema, self._out_schema = declobj.profiles[0]
        else:
            for i, o in declobj.profiles:
                if isinstance(i, NullSchema):
                    self._in_schema = i
                    self._out_schema = o
                    break
    
    @property
    def in_schema(self):
        return self._in_schema

    @property
    def out_schema(self):
        return self._out_schema

    def convert(self, value):
        u"""型変換処理。
        Caty のコマンドオブジェクトは複数のプロファイルを持ちうるため、
        このメソッドでは取りうるプロファイルについて総ざらいに変換処理を行い、
        成功した時点でプロファイルを確定する。
        """
        lasterror = None
        for i, o in self.declobj.profiles:
            try:
                v = i.convert(value)
                self._in_schema = i
                self._out_schema = o
                return v
            except Exception, e:
                raise

    def validate(self, value):
        u"""妥当性検証処理。
        基本的に convert と同様。
        """
        for i, o in self.declobj.profiles:
            try:
                i.validate(value)
                self._in_schema = i
                self._out_schema = o
                return
            except Exception, e:
                raise

    def parse_option(self, opts, args):
        u"""コマンドライン引数の解析。
        解析結果は辞書とリストのタプルであり、リストの値はスキーマによって変換された結果となる。
        boolean 値のトグルオプションと位置引数の曖昧さ回避のため、
        オプション（['foo', '1']という形式のリスト）と引数リストは
        別々に渡すこととする。
        """
        parser = self._make_option_parser()
        option, _ = parser.parse_args(opts)
        if option == {}: option = None
        assert _ == []
        return option, self.args_schema.convert(args)

    def _make_option_parser(self):
        parser = OptionParser()
        already_used = ['h', '-']
        if self.opts_schema.type.strip() in ('null', 'never', 'void'): return parser
        for k, v in self.opts_schema.iteritems():
            longopt = '--' + k
            shortopt = '-' + dropwhile(lambda x: x in already_used, list(k)).next()
            already_used.append(shortopt[1])
            type = v.type.strip()
            name = k if '-' not in k else k.replace('-', '_')
            remarks = v.remarks
            if type == 'number':
                type = 'float'
            elif type == 'integer':
                type = 'int'
            if type not in ('int', 'float', 'string', 'boolean'):
                raise Exception('Unsupported type: %s' % type)
            o = partial(parser.add_option, str(shortopt), str(longopt), dest=str(name))
            if type == 'boolean':
                o(action='store_true', default=False)
            else:
                o(action='store', type=type, default=None)
        return parser

    @property
    def facilities(self):
        u"""使用を宣言されたリソース(=Catyファシリティ)の名前を返す。
        """
        return self.declobj.get_all_resources()

    def get_command_class(self):
        return self.declobj.get_command_class()

    def resolve(self, module):
        for i, o in self.declobj.profiles:
            i.resolve(module)
            o.resolve(module)


    @property
    def usage(self):
        p = self._make_option_parser()
        opts = []
        for o in p.option_list:
            if o._long_opts[0] != '--help':
                if o.type:
                    opts.append(u'%s <%s>' % (o._long_opts[0], o.type))
                else:
                    opts.append(u'%s' % (o._long_opts[0]))

        return '\n'.join(opts), self.args_schema.type, self.in_schema.name, self.out_schema.name

