#coding: utf-8
from caty.util.web import *
from caty.core.facility import COMMIT
from caty.util.path import splitext, dirname, join
from caty.core.command.exception import *
from caty.core.exception import *
from caty.util import error_wrapper

import traceback

class AppDispatcher(object):
    u"""パス名に応じた Caty アプリケーションを返す。
    """
    def __init__(self, system):
        self._system = system

    def dispatch(self, path):
        name = path.split('/')[1].strip('/')
        if name in self._system.app_names:
            return self._system.get_app(name)
        else:
            return self._system.get_app('root')

class RequestHandler(object):
    u"""クライアントからのリクエストを編集し、実行可能 Caty コマンドを返す。
    クライアントからのリクエストは以下のように Caty が扱うには問題のあるものが含まれる。

    * ディレクトリ名にスラッシュなし
    * スラッシュの重複

    これらのリクエストを正規な形に修正した上で、 Caty コマンドを生成する。
    この時のクライアントは Web かコンソールかは関係なく、どちらも扱える。
    """

    def __init__(self, interpreter, 
                       env,
                       pub,
                       config,
                       encoding):
        self._app_path = env['APP_PATH']
        self._interpreter = interpreter
        self._verb_dispatcher = create_dispatcher(config['associations'], tuple(config['script_ext']))
        self._file = interpreter._facilities['pub']
        self._config = config
        self._encoding = encoding
        self._env = env

    def request(self, path, opts, method, transaction=COMMIT):
        path, vpath = self._create_path_and_vpath(path)
        self._env.put('PATH_INFO', vpath)
        self._env.put('REQUEST_METHOD', method)
        if '_verb' in opts:
            verb = opts['_verb']
            del opts['_verb']
        else:
            verb = ''
        return self._build(path, opts, verb, method, transaction)

    def _create_path_and_vpath(self, src):
        u"""実際にアクセスするファイルのパスとクライアントのアクセスしたパスのペアを返す。
        これは / で終わる URL など、実際に読み込まれるパスがクライアントのアクセスしたパスと異なる時があるためである。
        """
        full_path = unicode(src, self._encoding) if not isinstance(src, unicode) else src
        path = get_virtual_path('/'+self._app_path, full_path)
        vpath = path

        if path.endswith('/') and self._file.opendir(path).exists:
            for idf in self._config['indexfiles']:
                if self._file.open(path + idf).exists:
                    path += idf
                    break
        
        return path, vpath
    
    def _build(self, path, opts, verb, method, transaction):
        try:
            cmdstr = self._verb_dispatcher.get(self._file, path, verb, method)
            _f = self._file.opendir(path)
            if cmdstr is None and _f.exists:
                if (any(map(lambda f: self._file.open(join(_f.path, f)).exists, self._config['indexfiles']))):
                    cmd = self._interpreter.build(u'found %0', 
                                                  None, 
                                                  [join(self._env['HOST_URL'], self._app_path, path) + '/'], 
                                                  transaction=transaction)
                else:
                    raise IOError(path)
            else:
                cmd = self._interpreter.build(cmdstr, opts, [path], transaction=transaction)
        except Exception, e:
            return ExceptionAdaptor(e, traceback.format_exc())
        return PipelineAdaptor(cmd, self._interpreter.facilities['schema'])

class PipelineAdaptor(object):
    u"""コマンドを実行した後で適切なエラーハンドリングを行う。
    """
    def __init__(self, cmd, schema):
        self.__command = cmd
        self.__result = None
        self.__succeed = False
        self.__schema = schema

    @property
    def result(self):
        return self.__result

    @property
    def succeed(self):
        return self.__succeed

    @property
    def schema(self):
        return self.__schema

    def __call__(self, input):
        self.exec_cmd(input)
        return self.__result

    def _handle_pipeline(self, input, debug=False):
        try:
            result = self.__execute(input, debug)
            self.schema['Response'].validate(result)
            transaction = True
        except PipelineInterruption, e:
            result = e.json_obj
            transaction = True
        except PipelineErrorExit, e:
            result = e.json_obj
            transaction = False
        except SubCatyException, e:
            result = self._dispatch_error(e)
            transaction = False
        except CatyException, e:
            transaction = False
            result = self._make_error_response(e)
        except Exception, e:
            transaction = False
            result = {
                'status': 500,
                'body': u'Internal Server Error',
                'header': {
                    'content-type': u'text/plain; charset=utf-8',
                }
            }

        return result, transaction

    def __execute(self, input, debug):
        try:
            return self.__command(input)
        except Exception, e:
            if debug:
                import traceback
                traceback.print_exc()
            raise

    def exec_cmd(self, input, debug=False):
        result, transaction = self._handle_pipeline(input, debug)
        self.__result = result
        self.__succeed = transaction

    def _dispatch_error(self, e):
        if e.tag == 'FileNotFound':
            return {
                'status': 404,
                'header': {
                    'content-type': 'text/plain; charset=utf-8'
                },
                'body': e.get_message('ja')
            }
        else:
            return {
                'status': 500,
                'header': {
                    'content-type': 'text/plain; charset=utf-8'
                },
                'body': e.get_message('ja')
            }

    def _make_error_response(self, e):
        tag = e.tag
        value = e.error_obj
        if tag.startswith('HTTP_'):
            status = int(tag.replace('HTTP_', ''))
        else:
            status = 500
        result = {
            'status': status,
            'body': e.get_message('ja'),
            'header': {
                'content-type': u'text/html',
            }
        }
        return result

class ExceptionAdaptor(PipelineAdaptor):
    def __init__(self, error_obj, tb):
        self.__error_obj = error_obj
        self.__tb = tb
        PipelineAdaptor.__init__(self, None, None)

    def _handle_pipeline(self, input, debug=False):
        e = self.__error_obj
        if debug:
            print self.__tb
        if isinstance(e, SubCatyException):
            result = self._dispatch_error(e)
        elif isinstance(e, CatyException):
            result = self._make_error_response(e)
        else:
            result = {
                'status': 500,
                'body': u'Internal Server Error',
                'header': {
                    'content-type': u'text/plain; charset=utf-8',
                }
            }
        #print self.__tb
        return result, False

from caty.util import memoize
@memoize
def create_dispatcher(assocs, script_ext):
    return VerbDispatcher(assocs, script_ext)


class VerbDispatcher(object):
    def __init__(self, assocs, script_ext):
        self._ftd = FileTypeDispatcher(script_ext)
        for k, v in assocs.items():
            self._ftd.add(k, v)

    def get(self, fs, path, verb, method):
        return self._ftd.get(fs, path, verb, method)

class FileTypeDispatcher(object):
    u"""ファイル/ディレクトリに関連付けられた verb を返す。
    """
    def __init__(self, script_ext):
        self._entries = []
        self._entry_map = {} # filename.ext をサーチするための辞書
        self._script_ext = script_ext

    def add(self, type, value):
        if '|' in type:
            types = type.split('|')
        else:
            types = [type]
        for t in types:
            self._entries.append(VerbEntry(t, value))
            if '*' not in t and not t.startswith('.'):
                self._entry_map[t] = self._entries[-1]
        self._entries.sort()

    def get(self, fs, path, verb, method):
        ext = splitext(path)[1]
        if ext in self._script_ext:
            return fs.open(path).read()
        fname = path.split('/').pop(-1)
        if fname in self._entry_map:
            return self._entry_map[fname].get(fs, path, verb, method)
        for e in self._entries:
            r = e.get(fs, path, verb, method)
            if r:
                return r
        raise Exception(u'verb dispatch にマッチしませんでした。: %s' % path)


class VerbEntry(object):
    u"""verb dispatch の記述を解析し、 verb と method に応じたスクリプトを返す。
    verb dispatch の構文は以下の規則に従う。

    vd ::= verb ("/" method) ("#" "exists-parent")
    verb ::= 任意のアルファベット
    method ::= "GET" | "POST" | "PUT"

    """

    def __init__(self, pattern, verbs):
        self._pattern = pattern
        if pattern.endswith('/'):
            self._is_dir = True
            pattern = pattern[:-1]
        else:
            self._is_dir = False
        if '.' in pattern:
            self._basename, self._ext = splitext(pattern)
        else:
            self._basename = pattern
            self._ext = None
        self._verbs = {}
        for k, v in verbs.items():
            verb, method, parent = verb_parser.run(k)
            if not verb in self._verbs:
                self._verbs[verb] = {
                    method: {
                        'command': v,
                        'parent': parent
                    }
                }
            else:
                self._verbs[verb][method] = {
                    'command': v,
                    'parent': parent
                }

    def get(self, fs, path, verb, method):
        if self._is_dir:
            if not path.endswith('/'):
                f = fs.opendir(path+'/')
                if not f.exists:
                    return None
            path = path.rstrip('/')
        if not self._match(path):
            return None
        m = self._verbs.get(verb, None)
        if m is None:
            return None
            #raise Exception(u'未知の verb: %s/有効な verb: %s' % (verb, ', '.join(self._verbs.keys())))
        e = m.get(method.upper(), None)
        if e is None:
            if '__default__' in m:
                e = m['__default__']
            else:
                return None
            #raise Exception(u'未知の method: %s/有効な method %s' % (method, ', '.join(m.keys())))
        f = fs.opendir(path) if self._is_dir else fs.open(path)
        if not f.exists:
            if e['parent']:
                p = dirname(path) + '/' if dirname(path) != '/' else '/'
                if p == path: # ディレクトリが渡された場合
                    p = p.rstrip('/').rsplit('/', 1)[0]
                pdir = fs.opendir(p)
                if pdir.exists:
                    return e['command']
            raise FileNotFound(path)
        else:
            return e['command']

    def _match(self, path):
        return self._match_ext(path) and self._match_base(path)

    def _match_ext(self, path):
        tail = path.rstrip('/').rsplit('/', 1)[-1]
        if self._ext and self._ext != '.*': #ワイルドカードでない拡張子パターン
            if not tail.endswith(self._ext):
                return False
        elif self._ext == '.*': #ワイルドカード拡張子パターン
            return True
        else: #無拡張子パターン
            return '.' not in tail
        return True

    def _match_base(self, path):
        basename, _ = splitext(path.split('/').pop(-1))
        if self._basename in ('*', ''):
            return True
        else:
            return self._basename == basename

    def __cmp__(self, another):
        return cmp(self.score, another.score)

    @property
    def score(self):
        # verb dispatch の優先順位を決定する
        # ワイルドカードをパス名や拡張子に含むほど、スコアが低くなる
        s = 0
        if '.*' in self._pattern:
            s += 100
        if self._pattern.startswith('*'):
            s += 10
        if '.' not in self._pattern:
            s += 1
        return s

class Error(object):
    def get(self, fs, path, verb, method):
        raise IOError(path)

from topdown import *
@as_parser
def verb_parser(seq):
    v = seq.parse(option(verb, ''))
    m = seq.parse(option(method, 'GET'))
    e = bool(seq.parse(option(parent, False)))
    return v, m, e

def verb(seq):
    return seq.parse(Regex('[a-zA-Z0-9_-]*'))

def method(seq):
    return seq.parse(['/GET', '/POST', '/PUT', '/DELETE'])[1:]

def parent(seq):
    return seq.parse('#exists-parent')



