#coding: utf-8
import copy
from caty.core.resource import Resource
from caty.jsontools import TaggedValue, tag, tagged, untagged, obj2path
from caty.util import merge_dict, error_to_ustr, bind2nd, to_unicode
from caty.core.schema.errors import *
from pbc import *
import caty


class SchemaBase(Resource, PbcObject):
    u"""あらゆるスキーマクラスのベースクラス。
    スキーマオブジェクトは以下の振る舞いについて何らかの定義がなされていなければならない。

    * validate(v): 値の検証。実際には _validate をオーバーライドすることが殆どである。
    * convert(v): 値の変換。実際には _convert をオーバーライドすることが殆どである。
    * name: スキーマの名称。デフォルトでは type と同じ値を返す。
    * definition: スキーマの定義名。型変数を用いるスキーマの場合、 name<型変数名> の形になる。デフォルトでは name と同じ。
    * type: スキーマの型名。デフォルトではエラーなので必ずオーバーライドすること。
    * union(s): スキーマのユニオン。デフォルトのままで問題ないはずである。
    * intersect(s): スキーマのインターセクション。 object と型変数以外は未定義であるべきである。
    * update(s): スキーマの合成。 object と型変数以外は未定義であるべきである。
    * _clone(*args, **kwds): 新たなスキーマオプションと型変数へのスキーマ割り当てによるクローン。
    * apply(type_vars): 型変数の適用。型変数適用後のスキーマオブジェクトを返す。デフォルトではただ自身を返す。

    スキーマ固有のオプションは全てコンストラクタ、あるいは clone に対して辞書で渡される。
    また、このオプションは全て read-only なプロパティとして外部からアクセス可能であるべきである。

    スキーマの名称と型名と定義名の違いに注意すること。組み込み型は三つが一致するが、非組み込み型は異なる。
    例えば以下の場合、 restrictedString はスキーマの名称及び定義名であるが、型名は string であり、
    同様に list はスキーマの名称であるが定義名は list<_T> であり、型名は array である。

    {{{
    type restrictedString = string(maxLength=255);
    type list<T> = [T*];
    }}}

    ユーザ定義型はオプションを取ることができず、組み込み型は型変数を取る事ができない。
    ユーザ定義型も上記の restrictedString のように型変数を定義時に宣言していない場合、
    型変数を受け取る事ができない。

    また、スキーマオブジェクトは以下の規約に沿って動作しなければならない。

    * 引数無しでインスタンス化できる
    * 引数無しで _clone メソッドが呼べる
    * 引数無しの clone が呼ばれた場合、あるいは引数の値が全て空の場合単に自身を返す
    * インスタンス化された後は常に不変オブジェクトとして振る舞う

    """

    __properties__ = [
        'options',
        'optional',
        'remarks',
        'name',
        'definition',
        'type',
    ]

    def __init__(self, options=None):
        Resource.__init__(self)
        self._options = options if options else {}
        PbcObject.__init__(self)

    def optional():
        def _get(self):
            return self._options.get('optional', False)
        def _set(self, v):
            self._options['optional'] = True
        return _get, _set
    optional = property(*optional())
    
    def remarks():
        def _get(self):
            return self._options.get('remarks', False)
        def _set(self, v):
            self._options['remarks'] = True
        return _get, _set
    remarks = property(*remarks())

    def format():
        def _get(self):
            return self._options.get('format', u'')
        def _set(self, v):
            self._options['format'] = v
        return _get, _set
    format = property(*format())

    def _get_resource_type():
        return 'schema'

    @property
    def options(self):
        return copy.deepcopy(self._options)

    @property
    def name(self):
        return self.type

    @property
    def definition(self):
        return self.type

    def _get_canonical_name(self):
        return self.name

    @property
    def type(self):
        raise NotImplementedError()

    @property
    def type_vars(self):
        return []

    def __or__(self, schema):
        u"""Union 型スキーマの構築。
        本来 Python において | はビット単位の or 演算だが、
        この場合は Union の構築に用いる。
        a | b == UnionSchema(a, b) である。
        """
        return self.union(schema)

    def union(self, schema):
        if self.type == schema.type and '__variable__' != self.type:
            raise Exception(u'タグが排他になっていません: %s' % self.type)
        return UnionSchema(self, schema)

    def __and__(self, another):
        if self.type == 'never':
            return self.intersect(another)
        if another.type == 'never':
            return another.intersect(self)
        if self.type == 'any':
            return self.intersect(another)
        if another.type == 'any':
            return another.intersect(self)
        if self.type != another.type:
            if '__variable__' not in (self.type, another.type):
                opt = self.optional or another.optional
                return NeverSchema({'optional': opt})
                #raise Exception(u'異なる型同士の & 演算はできません: %s & %s' % (str(self.type), str(another.type)))
            else:
                return IntersectionSchema(self, another)
        n = self.intersect(another)
        if self.optional or another.optional:
            n.optional = True
        return n
    
    def intersect(self, schema):
        u"""スキーマ同士の intersection。
        必ず派生クラスではオーバーライドすること。
        """
        raise JsonSchemaError(u'%s と %s の & 演算は定義されていません。' % (str(self.type), str(schema.type)))

    def __lshift__(self, another):
        u"""引数のスキーマを型変更作用素として扱い、新たなスキーマを返す。
        実際の処理は update メソッドで行う。
        デフォルトでは単にエラーであり、型変数と object 型のみがオーバーライドすること。
        """
        return self.update(another)

    def update(self, another):
        raise JsonSchemaError(u'%s と %s の << 演算は定義されていません。' % (str(self.type), str(schema.type)))


    def convert(self, value):
        u"""convert メソッドの実装。
        実際の処理は _convert メソッドで行い、その結果を検証した結果を返すことになる。
        デフォルトの _convert の実装では何もせずに値をそのまま返す。
        """
        try:
            self.validate(value)
        except:
            try:
                r = self._convert(value)
            except JsonSchemaError, e:
                raise
            except Exception, e:
                raise JsonSchemaError(u'型変換時に予期せぬ例外が発生しました。: %s' % error_to_ustr(e))
        else:
            return value
        self.validate(r)
        return r

    def _convert(self, value):
        return value


    def clone(self, *args, **kwds):
        u"""スキーマのクローン。
        いわゆるプロトタイプパターン。
        """
        #if self._empty(args) and self._empty(kwds):
        if not args and not kwds:
            return self._deepcopy()
        return self._clone(*args, **kwds)

    def _empty(self, obj):
        if obj is None:
            return True
        if obj == []:
            return True
        if obj == {}:
            return True
        if isinstance(obj, (list, tuple)):
            return all([self._empty(x) for x in obj])
        if isinstance(obj, dict):
            return all([self._empty(x) for x in obj.values()])
        return False

    def _clone(self, *args, **kwds):
        raise NotImplementedError()

    def validate(self, value):
        u"""スキーマ処理のベース。
        タグ付きの値は基本的に共通でエラーとし、タグを扱う場合はこのメソッドをサブクラスでオーバーライドする。
        """
        if isinstance(value, TaggedValue):
            raise JsonSchemaError(u'タグ付きの値が渡されました: %s' % tag(value))
        self._validate(value)

    def to_tagged(self, tag):
        return TagSchema(tag, self)

    def apply(self, type_vars):
        return self

    def resolve(self, module):
        pass

    def has_attribute(self, k):
        o = self.__class__.__dict__.get(k, None)
        if o:
            if isinstance(o, SchemaAttribute):
                return True
        return False

    def pp(self):
        return 'type %s = %s' % (self.name, self.dump(0, []))

    ##############
    # 契約の宣言 #
    ##############

    if caty.DEBUG:
        def __invariant__(self):
            assert isinstance(self._options, dict)
    
        def _not_same_instance(self, result, old, *args, **kwds):
            assert self != result, (self, result)
    
        def _option_not_modified(self, result, old, *args, **kwds):
            assert self.options == old.options
    

        clone = Contract(clone)
        __and__ = Contract(__and__)
        __or__ = Contract(__or__)
        __lshift__ = Contract(__lshift__)
        __lshift__.ensure += _not_same_instance
        __or__.ensure += _not_same_instance
        __and__.ensure += _not_same_instance

class UnionSchema(SchemaBase):
    u"""union 型に対応するスキーマ。
    union 型は二つのスキーマによってインスタンス化され、
    それらのスキーマのいずれか一方に適合するデータであれば妥当なデータとみなす。
    両方のスキーマで妥当性検証に失敗したときのみがエラーである。
    """
    def __init__(self, a, b):
        self.__schema_a = a
        self.__schema_b = b
        SchemaBase.__init__(self)

    def validate(self, value):
        u"""union 型に対するスキーマ適用処理。
        """
        try:
            self.__schema_a.validate(value)
        except JsonSchemaError, e1:
            try:
                self.__schema_b.validate(value)
            except JsonSchemaError, e2:
                raise JsonSchemaError(error_to_ustr(e1) + ' / ' + error_to_ustr(e2))
            else:
                return
        else:
            return

    def intersect(self, another):
        if isinstance(another, UnionSchema):
            x, y = another.branch
            try:
                n = (x & self.__schema_a) | (y & self.__schema_b)
            except Exception, e:
                n = (y & self.__schema_a) | (x & self.__schema_b)
        else:
            n = (self.__schema_a & another) | (self.__schema_b & another)
        if self.optional or another.optional:
            n.optional = True
        return n

    def convert(self, value):
        try:
            return self.__schema_a.convert(value)
        except JsonSchemaError, e1:
            try:
                return self.__schema_b.convert(value)
            except JsonSchemaError, e2:
                raise JsonSchemaError(u'ユニオン型への変換に失敗しました。', e2.error_details)

    def dump(self, depth=0, node=[]):
        v = self.__schema_a.dump(depth, node) + ' | ' + self.__schema_b.dump(depth, node)
        return v

    def clone(self, options={}):
        s = self.__schema_a.clone() | self.__schema_b.clone()
        if self.optional or options.get('optional', False):
            s.optional = True
        return s

    def apply(self, type_vars):
        s = self.__schema_a.apply(type_vars) | self.__schema_b.apply(type_vars)
        if self.optional:
            s.optional = True
        return s

    @property
    def type(self):
        return '%s | %s' % (self.__schema_a.type, self.__schema_b.type)

    @property
    def name(self):
        return u'%s | %s' % (self.__schema_a.name, self.__schema_b.name)

    @property
    def definition(self):
        return u'%s | %s' % (self.__schema_a.definition, self.__schema_b.definition)

    @property
    def branch(self):
        return self.__schema_a, self.__schema_b

    if caty.DEBUG:
        def __invariant__(self):
            SchemaBase.__invariant__(self)
            self.__schema_a.__invariant__
            self.__schema_b.__invariant__

class IntersectionSchema(SchemaBase):
    u"""インターセクションスキーマ。
    通常、インターセクションは即座に計算されて IntersectionSchema でないスキーマになる。
    ただし、型変数を含めた演算は後の型解析のステップにならない限りは計算できない。
    実際にパイプラインが構築されて型解析が行われた場合、このスキーマは消滅して他のスキーマで置換される。
    """
    def __init__(self, a, b):
        assert '__variable__' in (a.type, b.type)
        self.__schema_a = a
        self.__schema_b = b
        SchemaBase.__init__(self)

    @property
    def type(self):
        return u'%s & %s' % (self.__schema_a.type, self.__schema_b.type)

    @property
    def name(self):
        return u'%s & %s' % (self.__schema_a.name, self.__schema_b.name)

    def clone(self, *args, **kwds):
        return self.__schema_a.clone() & self.__schema_b.clone()

    @property
    def definition(self):
        return u'%s & %s' % (self.__schema_a.definition, self.__schema_b.definition)

class UpdatorSchema(SchemaBase):
    u"""型アップデータスキーマ。
    インターセクション同様、型変数を含んだアップデート演算に対応するためのスキーマである。
    """
    def __init__(self, a, b):
        assert a.type == '__variable__'
        assert b.type != '__variable__'
        self.__schema_a = a
        self.__schema_b = b
        SchemaBase.__init__(self)

    @property
    def type(self):
        return u'%s << %s' % (self.__schema_a.type, self.__schema_b.type)

    @property
    def name(self):
        return u'%s << %s' % (self.__schema_a.name, self.__schema_b.name)

    @property
    def definition(self):
        return u'%s << %s' % (self.__schema_a.definition, self.__schema_b.definition)

    def clone(self, *args, **kwds):
        return self.__schema_a.clone() << self.__schema_b.clone()

_builtin_tags = [
    'integer',
    'number',
    'string',
    'binary',
    'boolean',
    'array',
    'object',
    'any',
    'null',
    'void',
    'never',
    'enum',
    ]

class TagSchema(SchemaBase):
    u"""タグ付き union で使われるスキーマ。
    通常のスキーマに tag プロパティが付いたものである。
    """
    def __init__(self, tag, schema):
        self.__schema = schema
        self._options = schema.options
        self.__tag = tag

    @property
    def _orig_schema(self):
        return self.__schema

    def clone(self, *args, **kwds):
        return TagSchema(self.__tag, self.__schema.clone(*args, **kwds))

    @property
    def tag(self):
        return self.__tag

    def validate(self, value):
        t = tag(value)
        if t != self.tag:
            if self.tag == '*' and t in _builtin_tags:
                raise JsonSchemaError(u'組み込み型はワイルドカードタグに適用できません: %s' % (t))
            if self.tag != '*':
                raise JsonSchemaError(u'タグが %s と一致していません: %s' % (self.tag, t))
        self.__schema.validate(untagged(value))

    def __and__(self, another):
        if not isinstance(another, TagSchema):
            raise JsonSchemaError(u'タグ %s と %s の & 演算は定義されていません' % (self.tag, another.type))
        if not (self.tag == '*' or another.tag == '*'):
            raise JsonSchemaError(u'タグ %s とタグ %s の & 演算は定義されていません' % (self.tag, another.tag))
        nt = self.tag if self.tag == '*' else another.tag
        return TagSchema(nt, self.__schema & another._orig_schema)

    def _convert(self, value):
        return tagged(self.__tag, self.__schema.convert(value))

    def dump(self, depth, node=[]):
        return '@%s %s' % (self.__tag, self.__schema.dump(depth, node))

    @property
    def type(self):
        return '@%s %s' % (self.__tag, self.__schema.type)

    @property
    def canonical_type(self):
        return '@%s %s' % (self.__tag, self.__schema.canonical_type)

class ScalarSchema(SchemaBase):

    def convert(self, value):
        u"""Web からの入力は常にリスト形式となる。
        そのため、1要素のリストは常にスカラー値に変換する。
        """
        if isinstance(value, list) and len(value) == 1:
            return SchemaBase.convert(self, value[0])
        else:
            return SchemaBase.convert(self, value)

    def _deepcopy(self):
        return self.__class__(self.options)

    def _clone(self, options):
        return self.__class__(options if options else self.options)

class AnySchema(SchemaBase):
    u"""あらゆる型を受け付けるスキーマ。
    つまり何もしない。
    """

    def intersect(self, another):
        return another.clone()

    def validate(self, value):
        pass

    def _convert(self, value):
        return value

    def dump(self, depth, node=[]):
        return 'any'


    def clone(self, *args, **kwds):
        return AnySchema(*args, **kwds)

    @property
    def type(self):
        return 'any'

class NullSchema(SchemaBase):
    u"""値が null 値、つまり None であれば良とするスキーマ。
    """
    def validate(self, value):
        if value is not None:
            raise JsonSchemaError(u'値が null ではありません')

    def intersect(self, another):
        if type(another) != NullSchema:
            raise JsonSchemaError(u'null と %s の間の & 演算は不可能です' % another.type)
        return NullSchema()

    def _convert(self, value):
        return value


    def clone(self, *args, **kwds):
        return NullSchema(*args, **kwds)

    def dump(self, depth, node=[]):
        return 'null'

    @property
    def type(self):
        return 'null'

class VoidSchema(SchemaBase):
    u"""Null と同じ unit だが、値なしを意味するところが違う。
    """


    def validate(self, value):
        pass

    def intersect(self, another):
        if type(another) != VoidSchema:
            raise JsonSchemaError(u'void と %s の間の & 演算は不可能です' % another.type)
        return VoidSchema()

    def _convert(self, value):
        return None

    def clone(self, *args, **kwds):
        return VoidSchema(*args, **kwds)

    def dump(self, depth, node=[]):
        return 'void'

    @property
    def type(self):
        return 'void'

class NeverSchema(SchemaBase):
    u"""そもそもアクセスがあってはいけないスキーマ。
    """
    def validate(self, value):
        raise JsonSchemaError('never')

    def intersect(self, another):
        return NeverSchema()

    def _convert(self, value):
        raise JsonSchemaError('never')

    def dump(self, depth, node=[]):
        return 'never'

    def clone(self, *args, **kwds):
        return NeverSchema(*args, **kwds)

    @property
    def type(self):
        return 'never'

class TypeVariable(AnySchema):
    def __init__(self, var_name):
        AnySchema.__init__(self)
        self.var_name = var_name

    def dump(self, depth, node=[]):
        return self.var_name

    def update(self, another):
        return UpdatorSchema(self, anothrt)

    def intersect(self, another):
        return IntersectionSchema(self, another)

    @property
    def name(self):
        return self.var_name

    @property
    def definition(self):
        return self.var_name

    @property
    def type(self):
        return '__variable__'

    def apply(self, vars):
        return vars[self.var_name]

    @property
    def type_vars(self):
        return [self.var_name]

    def clone(self, *args, **kwds):
        return TypeVariable(self.var_name)

class NamedSchema(SchemaBase):
    u"""名称付きスキーマ。
    組み込みでないスキーマは全て以下のように宣言される。

    {{{
    type list<T> = [T];
    type Person = {
        "name": string,
        "birth": integer
    };
    }}}

    これらはそれぞれ array, object と言ったスキーマであると同時に名称付きスキーマである。
    名称付きのスキーマは、型名の他に名称と定義名を値として持つ。
    名称とは list, Person のようなそのスキーマを参照する際に使う名前であり、
    定義名とは名称に型変数を加えたもの、即ち type 直後から = までの間に出現するトークンを指す。

    名称付きスキーマは型変数を持ちうる。上記の list の他、次のような tuple の定義も考えられる。

    {{{
    type tuple<X, Y> = [X, Y];
    type flipped<A, B> = tuple<B, A>; <- clone
    type x = flipped<integer, string>; 
    }}}

    二番目の例は単に要素の出現順序を逆転させただけで実用的な例ではないが、
    一応こういうことも可能となっている。

    """
    def __init__(self, name, type_vars, schema, module):
        SchemaBase.__init__(self)
        Resource.__init__(self, module.application)
        self._name = name
        self._type_vars = type_vars if type_vars else ()
        self._schema = schema
        self._module = module

    @property
    def canonical_name(self):
        an = self.application.name
        mn = self._module.name
        name = self.name
        if mn and mn != 'public':
            name = '%s:%s' % (mn, name)
        if an:
            name = '%s#%s' % (an, name)
        return name

    @property
    def name(self):
        return self._name
    
    @property
    def definition(self):
        return self._name if not self._type_vars else u'%s<%s>' % (self._name, ', '.join(self._type_vars))

    @property
    def type(self):
        return self._schema.type

    @property
    def type_vars(self):
        return self._type_vars

    def validate(self, value):
        self._schema.validate(value)
        if self._schema.format: # フォーマットの指定がある場合
            r = self._module.get_plugin(self._schema.format).validate(value)
            if r:
                raise JsonSchemaError(r)

    def convert(self, value):
        return self._schema.convert(value)

    def intersect(self, another):
        return self._schema & another

    def update(self, another):
        return self._schema << another

    def union(self, another):
        return self._schema | another

    def clone(self, *args, **kwds):
        return NamedSchema(self._name, 
                           self._type_vars, 
                           self._schema.clone(*args, **kwds), 
                           self._module)

    def __getattr__(self, name):
        if '_schema' in self.__dict__:
            return getattr(self._schema, name)
        else:
            raise AttributeError(name)

    def __getitem__(self, k):
        return self._schema[k]

    def __iter__(self):
        return iter(self._schema)

    def apply(self, vars):
        u"""型変数をスキーマに適用する。
        """
        return NamedSchema(self._name, 
                           self._type_vars, 
                           self._schema.apply(vars), 
                           self._module)

    def dump(self, depth, node):
        if self.name in node:
            return self.name
        else:
            return self._schema.dump(depth, node+[self.name])

    def optional():
        def _get(self):
            return self._schema.optional
        def _set(self, v):
            self._schema.optional = v
        return _get, _set
    optional = property(*optional())

    def __setattr__(self, n, v):
        if '_schema' in self.__dict__ and self._schema.has_attribute(n):
            setattr(self._schema, n, v)
        else:
            object.__setattr__(self, n, v)

class OverlayedDict(object):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __getitem__(self, k):
        if k in self.a:
            return self.a[k]
        elif k in self.b:
            return self.b[k]
        raise KeyError(k)

    def __contains__(self, k):
        return k in self.a or k in self.b

class SchemaAttribute(property):
    def __init__(self, name, default=None):
        property.__init__(self, lambda obj: obj.options.get(name, default), lambda obj, v: obj._options.__setitem__(name, v))

attribute = SchemaAttribute

