#coding: utf-8
from caty.core.schema.base import *

class PseudoTag(object):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def exclusive(self, another):
        return self.name == another.name and self.value != another.value

    def defined(self):
        return self.name

    def __str__(self):
        return '%s=%s' % (self.name, repr(self.value))

class ObjectSchema(SchemaBase):
    u"""JSON オブジェクトの妥当性検証を行うクラス。
    """
    maxProperties = attribute('maxProperties', -1)
    minProperties = attribute('minProperties', -1)
    pseudoTag = attribute('pseudoTag', PseudoTag(None, None))
    def __init__(self, schema_obj=None, wildcard=None, options=None):
        SchemaBase.__init__(self, options)
        self._wildcard = wildcard if wildcard else NeverSchema()
        self.schema_obj = schema_obj if schema_obj else {}
        assert all(map(lambda x:isinstance(x, SchemaBase), self.schema_obj.values())), repr(self.schema_obj.values())

    def wildcard():
        def _get(self):
            return self._wildcard

        def _set(self, v):
            self._wildcard = v

        return _get, _set
    wildcard = property(*wildcard())
    
    def union(self, schema):
        if schema.type == 'object':
            if not self.pseudoTag.exclusive(schema.pseudoTag):
                raise Exception(u'擬似タグが排他になっていません: %s, %s' % (str(self.pseudoTag), str(schema.pseudoTag)))
        return UnionSchema(self, schema)

    def intersect(self, another):
        cls = type(another)
        if not (cls in (UnionSchema, IntersectionSchema, UpdatorSchema, TypeVariable, ObjectSchema) 
                or cls == NamedSchema and another.type == 'object'):
            raise JsonSchemaError(u'%s との & 演算は定義されていません。' % (str(another.type)))
        if cls == ObjectSchema:
            return self._intersect_obj(another)
        elif cls == TypeVariable:
            return IntersectionSchema(self, another)
        return another.intersect(self)

    def _intersect_obj(self, another):
        newobj = ObjectSchema()
        if self.optional or another.optional:
            newobj.optional = True
        newschema_obj = merge_dict(self.schema_obj, another.schema_obj, lambda a, b: a & b)
        newobj.schema_obj = newschema_obj
        wild = self.wildcard & another.wildcard
        newobj.wildcard = wild
        ps1 = self.pseudoTag
        ps2 = self.pseudoTag
        if ps1.defined and ps2.defined:
            if ps1.exclusive(ps2):
                raise Exception(u'異なる擬似タグ同士の & 演算: %s, %s' % (str(self.pseudoTag), str(another.pseudoTag)))
            newobj.pseudoTag = ps1
        elif ps.defined and not ps2.defined:
            newobj.pseudoTag = ps1
        else:
            newobj.pseudoTag = ps2
        return newobj

    def update(self, another):
        if another.type != 'object' or (isinstance(another.type, tuple) and not all(map(lambda x:x.type == 'object', another.type))):
            raise JsonSchemaError('%s との << 演算は定義されていません。' % (str(another.type)))
        newobj = ObjectSchema()
        newschema_obj = merge_dict(self.schema_obj, another.schema_obj, 'error')
        newobj.schema_obj = newschema_obj
        return newobj

    def _validate(self, value):
        errors = {}
        is_error = False
        if not self.optional and value == None:
            raise JsonSchemaError(u'null は許可されていません', {})
        elif self.optional and value == None:
            return
        if not isinstance(value, dict):
            raise JsonSchemaError(u'値は JSON object (dict) でなければなりません', {})
        if self.minProperties > -1 and self.minProperties > len(value):
            raise JsonSchemaError(u'プロパティの数は %d 以上でなければなりません' % self.minProperties, {})
        if self.maxProperties > -1 and self.maxProperties < len(value):
            raise JsonSchemaError(u'プロパティの数は %d 以内でなければなりません' % self.maxProperties, {})
        for k, v in value.iteritems():
            if k not in self.schema_obj:
                if self.wildcard is None:
                    errors[k] = ErrorObj(True, u'', u'', u'未知のプロパティ: %s' % k)
                    is_error = True
                else:
                    try:
                        self.wildcard.validate(v)
                    except JsonSchemaError, e:
                        is_error = True
                        errors[k] = e
            else:
                if self.pseudoTag.name == k:
                    if self.pseudoTag.value != v:
                        errors[k] = ErrorObj(True, u'', u'', u'擬似タグの値に合致していません: %s' % k)
                        is_error = True
                else:
                    try:
                        self.schema_obj[k].validate(v)
                    except JsonSchemaError, e:
                        is_error = True
                        errors[k] = e
        # optional でないメンバーで漏れがないかチェック
        for k, v in self.schema_obj.iteritems():
            if not v.optional and k not in value:
                errors[k] = ErrorObj(True, u'', u'', u'プロパティが足りません: %s' % k)
                is_error = True
        if is_error:
            e = JsonSchemaErrorObject(u'object 型の検証に失敗しました')
            e.update(errors)
            raise e

    def _format(self, errors):
        r = []
        for k, v in errors:
            r.append(u'    %s: %s\n' % (k, v.msg))
        return ''.join(r)

    def _convert(self, value):
        result = {}
        errors = {}
        orig_info = {}
        is_error = False
        if not isinstance(value, dict):
            raise JsonSchemaError(u'値は JSON object (dict) でなければなりません')
        for k, v in value.iteritems():
            if k not in self.schema_obj:
                if self.wildcard is None:
                    errors[k] = ErrorObj(True, u'', u'', u'未知のプロパティ: %s' % k)
                    is_error = True
                else:
                    try:
                        result[k] = self.wildcard.convert(v)
                    except JsonSchemaError, e:
                        is_error = True
                        errors[k] = e
            else:
                try:
                    result[k] = self.schema_obj[k].convert(v)
                except JsonSchemaError, e:
                    is_error = True
                    errors[k] = e
        for k, v in self.schema_obj.iteritems():
            if not v.optional and k not in value:
                is_error = True
                errors[k] = ErrorObj(True, u'', u'', u'プロパティが足りません: %s' % k)
        if is_error:
            e = JsonSchemaErrorObject(u'object 型の検証に失敗しました')
            e.update(errors)
            e.succ = result
            raise e
        return result

    def __iter__(self):
        return iter(self.schema_obj.keys())

    def __getitem__(self, key):
        return self.schema_obj[key]

    def __setitem__(self, key, value):
        raise Exception(u'スキーマは不変オブジェクトです')
    
    def keys(self):
        return self.schema_obj.keys()

    def values(self):
        return self.schema_obj.values()

    def items(self):
        return self.schema_obj.items()

    def iterkeys(self):
        return self.schema_obj.iterkeys()

    def itervalues(self):
        return self.schema_obj.itervalues()

    def iteritems(self):
        return self.schema_obj.iteritems()

    def dump(self, depth=0, node=[]):
        s = []
        s.append('    ' * depth)
        s.append('{\n')
        depth += 1
        for k, v in self.schema_obj.iteritems():
            s.append('    ' * depth)
            s.append('"%s"' % k)
            s.append(':')
            s.append(v.dump(depth, node))
            if v.optional:
                s.append('?')
            s.append('\n')
        if self.wildcard:
            s.append('    ' * depth)
            s.append('*')
            s.append(':')
            s.append(self.wildcard.dump(depth, node))
            s.append('?')
            s.append('\n')
        else:
            s.append('    ' * depth)
            s.append('*:never')
            s.append('?')
            s.append('\n')
        s.append('    ' * (depth - 1))
        s.append('}')
        return ''.join(s)

    @property
    def type(self):
        return 'object'

    def apply(self, vars):
        d = {}
        for k, v in self.schema_obj.items():
            d[k] = v.apply(vars)
        wildcard = self.wildcard.apply(vars)
        return ObjectSchema(d, wildcard, self.options)

    def _deepcopy(self):
        return self._clone()

    def _clone(self, schema_obj=None, wildcard=None, options=None):
        if schema_obj:
            if all(map(lambda x:isinstance(x, SchemaBase), schema_obj.values())):
                return ObjectSchema(schema_obj, 
                                    wildcard if wildcard else self.wildcard,
                                    options if options else self.options)
        return ObjectSchema(self.schema_obj, 
                            wildcard if wildcard else self.wildcard,
                            options if options else self.options)

    def __eq__(self, o):
        return id(self) == id(o)

    def __ne__(self, o):
        return id(self) != id(o)

    @property
    def type_vars(self):
        v = []
        for s in self.schema_obj.values():
            for t in s.type_vars:
                if t not in v:
                    v.append(t)
        return v


    def __repr__(self):
        return object.__repr__(self)

    __str__ = __repr__


