# coding:utf-8
from caty.core.casm.finder import CommandFinder
from caty.core.facility import (FakeFacility, 
                                TransactionAdaptor, 
                                TransactionDiscardAdaptor, 
                                TransactionPendingAdaptor,
                                COMMIT,
                                ROLLBACK,
                                PEND)
from caty.core.script.parser import ScriptParser, NothingTodo
from caty.core.script.builder import CommandBuilder, CommandCombinator
from caty.core.script import node
from caty.command import builtin
from caty.core.command import Builtin, Syntax, Command, PipelineInterruption, PipelineErrorExit, ScriptError
from caty.util.cache import memoize, Cache
from caty.util import error_to_ustr
import types

def initialize(registrar, app, system):
    return ScriptModule(registrar, app, system)

class ScriptModule(object):
    def __init__(self, registrar, app, system):
        namespace = {}
        for n, profile_container in registrar.command_profiles:
            try:
                cmd = profile_container.get_command_class()
                cmd.profile_container = profile_container
                namespace[n] = cmd
            except Exception, e:
                raise
        self._finder = CommandFinder(namespace, app, system)
        self._enable_cache = system.enable_script_cache
        self._cache = Cache(1000)

    @property
    def enable_cache(self):
        return self._enable_cache

    @property
    def cache(self):
        return self._cache

    @property
    def finder(self):
        return self._finder

    def shell_mode(self, facilities):
        return ShellCommandInterpreter(facilities, self)

    def file_mode(self, facilities):
        return ScriptCommandInterpreter(facilities, self)

class AbstractCommandInterpreter(FakeFacility):
    u"""コマンド文字列からコマンドを構築するためのオブジェクト。
    実際の処理は CommandBuilder や parse_command に移譲し、
    このクラスでは対話シェルからの入力の受付や前二者の紐付けなどを行う。
    """
    def __init__(self, facilities, module):
        self.builder = CommandBuilder(facilities, module.finder)
        self._facilities = facilities
        self.prev_string = u''
        self.prev_opts = None
        self.prev_args = None
        self._cache = module.cache
        self.cache_enabled = module.enable_cache
        self._finder = module.finder
    
    def facilities():
        def get(self):
            return self._facilities

        def set(self, f):
            self._facilities = f
            self.builder = CommandBuilder(self._facilities, self._finder)
        return get, set
    facilities = property(*facilities())

    def _build_command(self, string, opts=None, args=None):
        proxy = None
        if self.cache_enabled:
            copts = () if not opts else tuple(opts.items())
            cargs = () if not args else tuple(args)
            key = (copts, cargs, string)
            proxy = self._cache.get(key)
        if not proxy:
            proxy = self._build(string, opts, args)
            if self.cache_enabled:
                self._cache.set(key, proxy)
        if proxy is not None:
            c = proxy.instantiate(self.builder)
            c.set_facility(self.facilities)
            return c
        else:
            return None

    def build(self, string, opts=None, args=None, transaction=COMMIT):
        u"""コマンド文字列とコマンドライン引数リストを受け取り、コマンドオブジェクトを構築する。
        """
        c = self._build_command(string, opts, args)
        if c:
            if transaction == COMMIT:
                return TransactionAdaptor(c, self.facilities)
            elif transaction == ROLLBACK:
                return TransactionDiscardAdaptor(c, self.facilities)
            elif transaction == PEND:
                return TransactionPendingAdaptor(c, self.facilities)
            else:
                assert False, u'トランザクションモードが不正です: %d' % (transaction)
        else:
            return c


    def _build(self, string):
        raise NotImplemtentedError
 
    def has_command(self, pkg, name):
        if pkg == 'builtin':
            return name in self.builder.namespace
        else:
            return pkg + ':' + name in self.builder.namespace

    def get_help(self, pkg, name):
        if pkg == 'builtin':
            return self.builder.namespace[name].profile_container.usage
        else:
            return self.builder.namespace[pkg + ':' + name].profile_container.usage
   
    def get_modules(self):
        s = set()
        for k in self.builder.namespace.keys():
            if ':' in k:
                s.add(k.rsplit(':', 1)[0])
        return sorted(list(s))

    def get_commands(self, name):
        for k, v in self.builder.namespace.items():
            if name != 'builtin' and k.startswith(name+':'):
                yield v.profile_container.name, v.profile_container.title
            elif name == 'builtin' and not ':' in k:
                if not issubclass(v, Syntax):
                    yield v.profile_container.name, v.profile_container.title

class ShellCommandInterpreter(AbstractCommandInterpreter):
    def _build(self, string, opts, args):
        u"""コマンド文字列を受け取り、中間形式のパイプラインを構築する。
        対話シェルから使われることを想定しているので、入力途中で改行されたりなど
        パイプラインが構築し切れていない場合は以前の状態を保存し、
        継続行の入力も受け付けるものとする。
        """
        from topdown import ParseFailed
        string = self.prev_string + '\n' + string if self.prev_string else string
        _opts = self.prev_opts if self.prev_opts else opts
        _args = self.prev_args if self.prev_args else args
        parser = ScriptParser(string, self.facilities, _opts, _args)
        try:
            pipeline = parser.parse()
        except NothingTodo:
            self.prev_string = u''
            raise
        except ParseFailed, e:
            self.prev_string = u''
            raise ScriptError(error_to_ustr(e))
        if not pipeline:
            self.prev_string = string
            self.prev_opts = opts
            self.prev_args = args
            return None
        else:
            self.prev_string = u''
            return pipeline

    def update_facility(self, facilities):
        self.builder.facilities = facilities
        self.facilities = facilities


class ScriptCommandInterpreter(AbstractCommandInterpreter):
    def _build(self, string, opts, args):
        u"""コマンド文字列を受け取り、中間形式のパイプラインを構築する。
        呼び出し可能な状態のオブジェクトが構築できなかった場合、即座にエラーとする。
        """
        parser = ScriptParser(string, self.facilities, opts, args)
        pipeline = parser.parse()
        if not pipeline:
            raise ScriptError(u'Syntax Error: %s' % string)
        else:
            return pipeline

def apply_builder(builder, cmd):
    u"""print コマンドは内部的にアソシエーションを解決する必要があるので、
    実行前に builder （実体は CommandInterpreter）を注入する。
    初期化時に行わないのは、オブジェクトの deepcopy 処理での不具合を回避するため。
    """
    if isinstance(cmd, (builtin.Print, builtin.Expand)):
        cmd.builder = builder
    elif isinstance(cmd, CommandCombinator):
        apply_builder(builder, cmd.bf)
        apply_builder(builder, cmd.af)
    return cmd


