#coding: utf-8
u"""Caty スクリプトノードモジュール。
Caty では switch, dispatch, object 生成などはすべてコマンドとして実装する。
"""

from caty.jsontools.path import build_query
from caty.jsontools import TaggedValue, tagged, untagged, TagOnly
from caty.jsontools import jstypes
from caty.core.command import ScriptError, PipelineInterruption, PipelineErrorExit, Command, Syntax
import types

schema = ''

class ScalarBuilder(Syntax):
    command_decl = u"""command scalar-builder :: void -> _T
                        reads schema
                        refers python:caty.core.script.node.ScalarBuilder;
    """

    def __init__(self):
        Syntax.__init__(self)

    def set_value(self, value):
        if isinstance(value, str):
            v = unicode(value)
        else:
            v = value
        #self.profile.out_schema = jstypes.enum.create({}, [v])
        self.value = v

    def execute(self):
        return self.value

class ListBuilder(Syntax):
    command_decl = u"""command list-builder :: T -> array
                        refers python:caty.core.script.node.ListBuilder;
    """

    def __init__(self):
        Syntax.__init__(self)
        self.values= []

    def set_values(self, values):
        scms = []
        for v in values:
            scms.append(v.out_schema)
            if isinstance(v, Command):
                self.values.append(v)
            else:
                s = ScalarBuilder()
                s.set_value(v)
                self.values.append(s)
        self.out_schema.schema_list = scms

    def set_facility(self, facilities):
        for v in self.values:
            v.set_facility(facilities)

    def execute(self, input):
        r = []
        for v in self.values:
            r.append(v(input))
        return r


class ObjectBuilder(Syntax):
    command_decl = u"""command object-builder :: T -> object
                        refers python:caty.core.script.node.ObjectBuilder;
    """

    def __init__(self):
        Syntax.__init__(self)
        self.__nodes = {}

    def add_node(self, node):
        if node.name in self.__nodes:
            raise KeyError(node.name)
        self.__nodes[node.name] = node

    def set_facility(self, facilities):
        for n in self.__nodes.values():
            n.set_facility(facilities)

    def execute(self, input):
        obj = {}
        for name, node in self.__nodes.iteritems():
            obj[name] = node(input)
        if '$tag' in obj:
            if '$val' in obj:
                o = tagged(obj['$tag'], obj['$val'])
            else:
                t = obj['$tag']
                del obj['$tag']
                o = tagged(t, obj)
            return o
        else:
            return obj


class ConstNode(object):
    def __init__(self, name, value):
        self.name = name
        self.cmd = ScalarBuilder()
        self.cmd.set_value(value)

    def set_facility(self, facilities):
        self.name = self.name if isinstance(self.name, unicode) else unicode(self.name, facilities['env'].read_mode.get('SYSTEM_ENCODING'))

    def __call__(self, input):
        return self.cmd(input)

class CommandNode(object):
    def __init__(self, name, cmd):
        self.name = name
        self.cmd = cmd

    def set_facility(self, facilities):
        self.cmd.set_facility(facilities)
        self.name = self.name if isinstance(self.name, unicode) else unicode(self.name, facilities['env'].read_mode.get('SYSTEM_ENCODING'))

    def __call__(self, input):
        return self.cmd(input)

class Dispatch(Syntax):
    command_decl = u"""command dispatch :: any -> any
                        refers python:caty.core.script.node.Dispatch;
    """

    def __init__(self):
        Syntax.__init__(self)
        self.__cases = {}
        self.__query = build_query('$')
        self.__scalar_tag_map = {
            types.IntType: ['integer', 'number'],
            types.FloatType: ['number'],
            types.UnicodeType: ['string'], 
            types.BooleanType: ['boolean'], 
            types.NoneType: ['null'], 
            types.ListType: ['array'],
            types.DictType: ['object'],
        }
    
    def add_case(self, case):
        self.__cases[case.tag] = case

    def set_facility(self, facilities):
        for k ,v in self.__cases.items():
            v.set_facility(facilities)

    def execute(self, input):
        jsobj = input
        target = self.__query.find(jsobj).next()
        if not isinstance(target, TaggedValue) and not (isinstance(target, dict) and '$tag' in target):
            return self.not_tagged_value_case(target)
        else:
            return self.tagged_value_case(target)

    def tagged_value_case(self, target):
        tag = target.tag if isinstance(target, TaggedValue) else target['$tag']
        if tag not in self.__cases:
            raise ScriptError(tag)
        return self.exec_cmd(tag, target)

    def not_tagged_value_case(self, target):
        tags = self.__scalar_tag_map.get(type(target), None)
        if tags == None:
            raise ScriptError()
        for tag in tags:
            if tag in self.__cases:
                return self.exec_cmd(tag, target)
        raise ScriptError(tag)
    
    def exec_cmd(self, tag, jsobj):
        childcmd = self.__cases[tag].cmd
        if childcmd.in_schema != jstypes.never:
            if isinstance(self.__cases[tag], UntagCase):
                input = untagged(jsobj)
            else:
                input = jsobj
        return childcmd(input)

class TagBuilder(Syntax):
    command_decl = u"""command add-tag :: any -> any
                        refers python:caty.core.script.node.TagBuilder;
    """
    def __init__(self, tag, cmd):
        Syntax.__init__(self)
        self.command = cmd
        self.tag = tag
    
    def set_facility(self, facilities):
        self.command.set_facility(facilities)

    def execute(self, input):
        return tagged(self.tag, self.command(input))

class UnaryTagBuilder(Syntax):
    command_decl = u"""command unary-tag :: any -> any
                        refers python:caty.core.script.node.UnaryTagBuilder;
    """
    def __init__(self, tag):
        Syntax.__init__(self)
        self.tag = tag
    
    def set_facility(self, facilities):
        pass

    def execute(self, input):
        return TagOnly(self.tag)


class Case(object):
    def __init__(self, tag, cmd):
        self.tag= tag
        self.cmd = cmd

    def set_facility(self, facilities):
        self.cmd.set_facility(facilities)

class UntagCase(Case):pass

class Each(Syntax):
    command_decl = u"""command each-functor-applied :: list <T> -> list <T>
                        refers python:caty.core.script.node.Each;"""
    def __init__(self, cmd):
        Syntax.__init__(self)
        self.cmd = cmd

    def set_facility(self, facilities):
        self.cmd.set_facility(facilities)
        ils = facilities['schema']['list']
        ils.schemata = [self.cmd.in_schema]
        ols = facilities['schema']['list']
        ols.schemata = [self.cmd.out_schema]
        self.profile.declobj.profiles = [(ils, ols)]

    def execute(self, input):
        r = []
        for v in input:
            r.append(self.cmd(v))
        return r

from caty.util import error_to_ustr
class Capture(Syntax):
    command_decl = u"""@[test] command captured-pipeline :: _S -> @OK _T | @Err _U
                            reads env
                        refers python:caty.core.script.node.Capture;"""
    def __init__(self, cmd):
        Syntax.__init__(self)
        self.cmd = cmd

    def set_facility(self, facilities):
        Syntax.set_facility(self, facilities)
        self.cmd.set_facility(facilities)

    def execute(self, input):
        try:
            return tagged('OK', self.cmd(input))
        except (PipelineErrorExit, PipelineInterruption), e:
            return tagged('Err', e.json_obj)
        except Exception, e:
            return tagged('Err', error_to_ustr(e))
    

class TypeInfo(Syntax):
    command_decl = u"""@[console] command type-info :: void -> string
                        reads schema
                        refers python:caty.core.script.node.TypeInfo;"""

    def __init__(self, cmd):
        Syntax.__init__(self)
        self.cmd = cmd

    def set_facility(self, facilities):
        Syntax.set_facility(self, facilities)
        self.cmd.set_facility(facilities)

    def execute(self):
        return unicode('%s -> %s' % (self.cmd.profile.in_schema.canonical_name, self.cmd.profile.out_schema.canonical_name))

class Time(Syntax):
    command_decl = u"""
    command time-functor :: T -> T
        refers python:caty.core.script.node.Time;
    """
    
    def __init__(self, cmd):
        Syntax.__init__(self)
        self.cmd = cmd

    def set_facility(self, facilities):
        Syntax.set_facility(self, facilities)
        self.cmd.set_facility(facilities)

    def execute(self, input):
        import time
        s = time.time()
        r = self.cmd(input)
        e = time.time()
        print e - s, '[sec]'
        return r

