#coding:utf-8
from __future__ import with_statement
from decimal import Decimal
from topdown import *
from caty.core.command import *
from caty.core.script.proxy import CommandProxy
from caty.core.script.proxy import ScalarProxy as ScalarBuilder
from caty.core.script.proxy import ListProxy as ListBuilder
from caty.core.script.proxy import ObjectProxy as ObjectBuilder
from caty.core.script.proxy import ConstNodeProxy as ConstNode
from caty.core.script.proxy import CommandNodeProxy as CommandNode
from caty.core.script.proxy import DispatchProxy as Dispatch
from caty.core.script.proxy import TagProxy as TagBuilder
from caty.core.script.proxy import UnaryTagProxy as UnaryTagBuilder
from caty.core.script.proxy import CaseProxy as Case
from caty.core.script.proxy import UntagCaseProxy as UntagCase
from caty.core.script.proxy import EachProxy as Each
from caty.core.script.proxy import TimeProxy as Time
from caty.core.script.proxy import CaptureProxy as Capture
from caty.core.script.proxy import TypeInfoProxy as TypeInfo
from caty.core.script.proxy import combine_proxy
from caty.util import bind2nd
import caty.jsontools.xjson as xjson
from caty.core.exception import SubCatyException

class NothingTodo(Exception):
    u"""コメントのみの入力など、何もしないときのシグナル
    """
    def __init__(self):
        pass

class HashNotationFound(SubCatyException):
    error_type = 'hash-notation'
    message_map = {
        'ja': u'ハッシュ記法は Caty スクリプトとして評価できません',
        'en': u'The hash-notation is not able to evaluate as an Caty script'
    }

class ScriptParser(object):
    def __init__(self, text, facilities, opts=None, args=None):
        self.text = text
        self.mafs = facilities['pub'].read_mode
        scripts = facilities.get('scripts', None)
        if scripts:
            self.scripts = scripts
        self.facilities = facilities
        self.opts = opts if opts else {}
        self.args = args if args else []

    def parse(self):
        if not self.text:
            return None
        seq = CharSeq(self.text, hook=xjson.remove_comment, auto_remove_ws=True)
        try:
            self._script = self.make_pipeline(seq)
            if not seq.eof:
                raise ParseFailed(seq, self.make_pipeline)
        except EndOfBuffer, e:
            return None
        return self._script

    def value(self, cs):
        v = cs.parse([xjson.string, 
                      xjson.multiline_string,
                      bind2nd(xjson.null, True), 
                      bind2nd(xjson.number, True), 
                      bind2nd(xjson.integer, True), 
                      bind2nd(xjson.boolean, True),
                      ])
        s = ScalarBuilder()
        s.set_value(v)
        return s

    def name(self, seq):
        return seq.parse(Regex(r'[a-zA-Z]+[-a-zA-Z0-9:._]*'))

    def functor(self, seq):
        func = seq.parse(['each', 'capture', 'type-info', 'time'])
        seq.parse('{')
        cmd  = self.make_pipeline(seq)
        seq.parse('}')
        if func == 'each':
            return Each(cmd)  
        elif func == 'type-info':
            return TypeInfo(cmd)
        elif func == 'time':
            return Time(cmd)
        else:
            return Capture(cmd)

    def command(self, seq):
        name = self.name(seq)
        opts = self.options(seq)
        args = self.arguments(seq)
        return CommandProxy(name, opts, args)

    def arguments(self, seq):
        args = filter(lambda a: a is not None, seq.parse(many(self.arg)))
        if not args:
            args = None
        return args

    def options(self, seq):
        opts = filter(lambda o: o != (None, None), seq.parse(many(self.opt)))
        o = []
        if opts:
            for op, val in opts:
                o.append(op)
                if val:
                    o.append(val)
        else:
            o = []
        return o

    def pipe(self, seq):
        _ = seq.parse(['|','>@'])
        return _

    def pipeline(self, seq):
        r = []
        seq.parse(self.do_nothing)
        if seq.eof:
            raise NothingTodo()
        r.append(seq.parse([self.term, self.group]))
        while True:
            a = option(self.pipe, None)(seq)
            if not a:
                break
            if a == '|':
                r.append(seq.parse(self.term))
            elif a == '>@':
                t = self.tag_name(seq)
                exp = r.pop(-1)
                r.append(TagBuilder(t, exp))
        return combine_proxy(r)

    def group(self, seq):
        _ = seq.parse('(')
        r = self.pipeline(seq)
        _ = seq.parse(')')
        return r

    def make_pipeline(self, seq):
        return seq.parse([self.pipeline])

    def term(self, seq):
        parsers = map(try_, [
                    self.functor,
                    self.tag,
                    self.cond,
                    self.object, 
                    self.diagram_order, 
                    self.value, 
                    self.script,
                    self.command,
                    self.list, 
                    self.hash,
                    ])
        return seq.parse(parsers)

    def tag(self, seq):
        _ = seq.parse('@')
        n = self.tag_name(seq)
        delim = True if seq.eof else seq.peek(option([',', ']', '}'], False))
        if delim:
            return UnaryTagBuilder(n)
        r = seq.parse([self.term, self.group])
        return TagBuilder(n, r)
    
    def tag_name(self, seq):
        return seq.parse([xjson.string, self.name, Regex(r'[-0-9a-zA-Z_]+')])

    def unquoted(self, seq):
        return seq.parse(Regex(r'[^\t\r\n <>|%+"(){},\[\]]+'))

    def opt(self, seq):
        seq.ignore_hook = True
        try:
            o, v = seq.parse([self.shortopt, self.longopt, self.parameter_opt])
            skip_ws(seq)
            return o, v
        finally:
            seq.ignore_hook = False

    def script(self, seq):
        path = seq.parse(Regex(r'[^\t\r\n <>|+"(){},\[\]]+\.(caty|cgi)'))
        s = None
        if path.startswith('/'):
            with self.mafs.open(path) as f:
                if f.exists:
                    s = f.read()
        elif self.scripts:
            with self.scripts.open('/' + path) as f:
                if f.exists:
                    s = f.read()
        if not s:
            raise ParseFailed(seq, self.script)
        opts = self.options(seq)
        args = self.arguments(seq)
        if not args:
            args = [""]
        else:
            args = [""] + args
        sub_parser = ScriptParser(s, self.facilities, opts, args)
        try:
            p = sub_parser.parse()
        except ParseFailed, e:
            raise e
        return p

    def shortopt(self, seq):
        o = seq.parse(Regex(r'-[a-zA-Z]'))
        try:
            v = seq.parse([xjson.number, xjson.integer, self.unquoted, xjson.string])
            return o, v
        except:
            return o, None

    def longopt(self, seq):
        o = seq.parse(Regex(r'--[a-zA-Z]+[-a-zA-Z0-9_]*'))
        try:
            seq.parse('=')
            v = seq.parse([xjson.number, xjson.integer, self.unquoted, xjson.string, self.named_arg, self.indexed_arg])
            if v is None: return (None, None)
            return o, v
        except:
            return o, None

    def parameter_opt(self, seq):
        name = seq.parse(Regex(r'%--[a-zA-Z]+[-a-zA-Z0-9_]*\??'))[3:]
        optional = False
        if name.endswith('?'):
            name = name[:-1]
            optional = True
        if optional:
            if name in self.opts:
                return '--' + name, self.opts[name]
            else:
                return (None, None)
        else:
            if name in self.opts:
                return '--' + name, self.opts[name]
            else:
                raise ParseError(seq, self.parameter_opt)
    
    def named_arg(self, seq):
        name = seq.parse(Regex(r'%[a-zA-Z]+[-a-zA-Z0-9_]*\??'))[1:]
        optional = False
        if name.endswith('?'):
            name = name[:-1]
            optional = True
        if optional:
            if name in self.opts:
                return self.opts[name]
            else:
                return None
        else:
            if name in self.opts:
                return self.opts[name]
            else:
                raise ParseError(seq, self.named_arg)
 
    def arg(self, seq):
        r = seq.parse([xjson.number, xjson.integer, xjson.boolean, self.unquoted, xjson.string, self.named_arg, self.indexed_arg])
        return r

    def indexed_arg(self, seq):
        index = seq.parse(Regex(r'%[0-9]+\??'))[1:]
        optional = False
        if index.endswith('?'):
            index = int(index[:-1])
            optional = True
        else:
            index = int(index)
        if len(self.args) > index:
            return self.args[index]
        else:
            return None

    def comma(self, seq):
        _ = seq.parse(',')
        return _

    def list(self, seq):
        seq.parse('[')
        values = seq.parse(split(self.listitem, self.comma, True))
        seq.parse(']')
        l = ListBuilder()
        l.set_values(anything(values))
        return l

    def listitem(self, seq):
        if seq.current == ']': return
        try:
            return self.make_pipeline(seq)
        except NothingTodo:
            raise ParseFailed(seq, self.listitem)

    def object(self, seq):
        seq.parse('{')
        items = seq.parse(split(self.item, self.comma, True))
        seq.parse('}')
        o = ObjectBuilder()
        for i in anything(items):
            o.add_node(i)
        return o

    def item(self, seq):
        if seq.current == '}': return
        n = xjson.string(seq)
        seq.parse(':')
        try:
            v = self.make_pipeline(seq)
        except NothingTodo:
            raise ParseFailed(seq, self.item)
        return CommandNode(n, v)

    def diagram_order(self, seq):
        seq.parse('do')
        seq.parse('{')
        items = seq.parse(split(self.item_do, self.comma))
        seq.parse('}')
        o = ObjectBuilder()
        for i in anything(items):
            o.add_node(i)
        return o

    def item_do(self, seq):
        if seq.current == '}': return
        try:
            v = self.make_pipeline(seq)
        except NothingTodo:
            raise ParseFailed(seq, self.item)
        seq.parse('>:')
        n = xjson.string(seq)
        return CommandNode(n, v)

    def cond(self, seq):
        seq.parse('when')
        seq.parse('{')
        cases = seq.parse(split(self.case, self.comma, allow_last_delim=True))
        seq.parse('}')
        d = Dispatch()
        for c in anything(cases):
            d.add_case(c)
        return d

    def case(self, seq):
        t = seq.parse(self.tag_name)
        tp = seq.parse(['==>', '=>'])
        try:
            v = self.make_pipeline(seq)
        except NothingTodo:
            raise ParseFailed(seq, self.case)
        return Case(t, v) if tp == '==>' else UntagCase(t, v)

    def remove_comment(self, seq):
        seq.parse(self.singleline_comment)
        seq.parse(self.multiline_comment)
        seq.parse(skip_ws)

    def singleline_comment(self, seq):
        if seq.parse(option('//', None)) == None: return
        seq.parse([until('\n'), until('\r')])

    def multiline_comment(self, seq):
        if seq.parse(option('/*', None)) == None: return
        seq.parse(until('*/'))
        seq.parse('*/')

    def do_nothing(self, seq):
        u"""コメント行のみ入力された場合用のダミー
        """
        pass

    def hash(self, seq):
        seq.parse('#')
        seq.parse('(')
        name = seq.parse(Regex(r'[a-zA-Z_]+[a-zA-Z0-9_-]*'))
        parsers = map(try_, [
                    self.functor,
                    self.tag,
                    self.cond,
                    self.object, 
                    self.diagram_order, 
                    self.value, 
                    self.script,
                    self.command,
                    self.list, 
                    ])
        obj = seq.parse(option(parsers))
        seq.parse(')')
        raise HashNotationFound([])

def anything(l):
    return filter(lambda i: i!= None, l)

