#coding:utf-8
from caty.core.command import *
from caty.core.script.node import *
from caty.core.script.builder import CommandCombinator

class Proxy(object):
    u"""パイプライン構築のためのプロキシ。
    テスタビリティの確保と構文エラー発見容易性のために、
    パイプラインの構築の際にはプロキシクラスを間に挟み、遅延初期化を行う。
    基本的に caty.shell.script.node モジュールのクラスとプロキシは一対一対応する。
    プロキシと本体はコマンドオブジェクト独自のメソッド・プロパティを除いては
    ソースコード上で見えるメソッド・プロパティが一致している。
    """
    def instantiate(self, builder):
        raise NotImplemtentedError

class CommandProxy(Proxy):
    u"""コマンド呼び出しに遭遇したときに構築されるプロキシクラス。
    """
    def __init__(self, name, opts, args):
        self.name = name
        self.opts = opts
        self.args = args

    def instantiate(self, builder):
        return builder.build(self.name, self.opts, self.args)

class ScalarProxy(Proxy):
    def __init__(self):
        self.value = None

    def __call__(self):
        import traceback
        for l in traceback.format_stack(): print l

    def set_value(self, v):
        self.value = v

    def instantiate(self, builder):
        s = ScalarBuilder()
        s.set_value(self.value)
        return s

class ListProxy(Proxy):
    def __init__(self):
        self.values= []
    
    def set_values(self, values):
        self.values = values

    def instantiate(self, builder):
        l = ListBuilder()
        l.set_values([v.instantiate(builder) for v in self.values])
        return l

class ObjectProxy(Proxy):
    def __init__(self):
        self.nodes = []

    def add_node(self, node):
        self.nodes.append(node)

    def instantiate(self, builder):
        o = ObjectBuilder()
        for n in self.nodes:
            o.add_node(n.instantiate(builder))
        return o

class ConstNodeProxy(Proxy):
    def __init__(self, n, v):
        self.name = n
        self.value = v

    def instantiate(self, builder):
        return ConstNode(self.name, self.value)

class CommandNodeProxy(Proxy):
    def __init__(self, n, c):
        self.name = n
        self.cmdproxy = c

    def instantiate(self, builder):
        return CommandNode(self.name, self.cmdproxy.instantiate(builder))

class DispatchProxy(Proxy):
    def __init__(self):
        self.cases = []

    def add_case(self, case):
        self.cases.append(case)

    def instantiate(self, builder):
        d = Dispatch()
        for c in self.cases:
            d.add_case(c.instantiate(builder))
        return d

class TagProxy(Proxy):
    def __init__(self, t, c):
        self.tag = t
        self.cmdproxy = c

    def instantiate(self, builder):
        return TagBuilder(self.tag, self.cmdproxy.instantiate(builder))

class UnaryTagProxy(Proxy):
    def __init__(self, t):
        self.tag = t

    def instantiate(self, builder):
        return UnaryTagBuilder(self.tag)

class CaseProxy(TagProxy):
    def instantiate(self, builder):
        return Case(self.tag, self.cmdproxy.instantiate(builder))

class UntagCaseProxy(TagProxy):
    def instantiate(self, builder):
        return UntagCase(self.tag, self.cmdproxy.instantiate(builder))

class FunctorProxy(Proxy):
    def __init__(self, c):
        self.cmdproxy = c

    def instantiate(self, builder):
        r = self._get_class()(self.cmdproxy.instantiate(builder))
        builder.set_facility_to(r)
        return r

class EachProxy(FunctorProxy):
    def _get_class(self):
        return Each

class TimeProxy(FunctorProxy):
    def _get_class(self):
        return Time

class CaptureProxy(FunctorProxy):
    def _get_class(self):
        return Capture

class TypeInfoProxy(FunctorProxy):
    def _get_class(self):
        return TypeInfo

class CombinatorProxy(Proxy):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def instantiate(self, builder):
        return CommandCombinator(self.a.instantiate(builder), self.b.instantiate(builder))



def combine_proxy(args):
    return reduce(CombinatorProxy, args)


