# coding: utf-8
u"""JSON 関連のユーティリティ。
Python のバージョン間で JSON ライブラリの有無が違うことへの対処、
JSON オブジェクトへのスキーマ適用などの処理などを行う。
"""

try:
    import json
except:
    try:
        import simplejson as json
    except:
        raise ImportError('Python 2.6 (or later) or simplejson package is needed')

import decimal
__all__ = ['load', 
           'loads', 
           'dump', 
           'dumps', 
           'prettyprint', 
           'pp', 
           'TaggedValue',
           'TagOnly',
           'tag',
           'obj2path',
           'path2obj',
           'tagged',
           'untagged',
           'split_tag',
           'CatyEncoder',
           'raw_json']

raw_json = json

class JsonError(Exception): pass

from caty.util import deprecated
def load(fo):
    u"""ファイル類似オブジェクトより JSON データをロードする。
    """
    return decode(json.load(fo, parse_float=decimal.Decimal))

@deprecated(u'stdjsonかxjsonを使うこと')
def loads(s):
    u"""文字列より JSON データをロードする。
    """
    return decode(json.loads(s, parse_float=decimal.Decimal))

@deprecated(u'stdjsonかxjsonを使うこと')
def dump(obj, fo, **kwds):
    return json.dump(encode(obj), fo, cls=CatyEncoder, **kwds)

@deprecated(u'stdjsonかxjsonを使うこと')
def dumps(obj, **kwds):
    return json.dumps(encode(obj), cls=CatyEncoder, **kwds)

class CatyEncoder(json.encoder.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return InternalDecimal(o)
        else:
            return json.encoder.JSONEncoder.default(self, o)

class InternalDecimal(float):
    u"""simplejson の JSONEncoder は、 default をオーバーライドした結果をさらにエンコードするようになっている。
    そのため、 Decimal 型をエンコードしようとするときに default メソッド内部で文字列にすると、
    その結果を文字列としてさらにエンコードされるため、デコード処理の際に Decimal ではなく文字列とされてしまう。
    JSONEncoder に Decimal のエンコードをさせる事ができる。
    """
    def __init__(self, v):
        self.__value = v
        
    def __repr__(self):
        return str(self.__value)

def encode(obj):
    u"""Caty 内部形式から JSON 標準形式へと変換する。
    """
    if isinstance(obj, TaggedValue):
        return {'$tag': obj.tag, '$val': encode(obj.value)}
    elif isinstance(obj, TagOnly):
        return {'$tag': obj.tag, '$no-value': True}
    elif isinstance(obj, dict):
        n = {}
        for k, v in obj.items():
            n[k] = encode(v)
        return n
    elif isinstance(obj, (list, tuple)):
        return map(encode, obj)
    else:
        return obj

def decode(obj):
    u"""標準の JSON 形式から Caty 内部形式へと変換する。
    内部形式ではタグの扱いが異なる。
    """
    if isinstance(obj, dict):
        if '$tag' in obj:
            if '$val' in obj:
                return TaggedValue(obj['$tag'], decode(obj['$val']))
            elif '$no-value' in obj:
                return TagOnly(obj['$tag'])
            else:
                raise JsonError(u'$tag に対応する $val あるいは $no-value がありません')
        else:
            n = {}
            for k, v in obj.items():
                n[k] = decode(v)
            return n
    elif isinstance(obj, (list, tuple)):
        return map(decode, obj)
    else:
        if isinstance(obj, str):
            return unicode(obj)
        else:
            return obj

def prettyprint(obj, depth=0, encoding=None):
    def format(o, d):
        if isinstance(o, TaggedValue):
            return '@%s %s' % (tag(o), prettyprint(untagged(o), encoding=encoding))
        elif isinstance(o, TagOnly):
            return '@%s' % tag(o)
        elif isinstance(o, unicode):
            if encoding:
                o = o.encode(encoding)
            if '\r' in o or '\n' in o:
                return u"'''%s'''" % o
            else:
                return u'"%s"' % o.replace('\\', '\\\\').replace('"', '\\"')
        elif isinstance(o, list):
            return u'[%s]' % (', '.join(map(lambda _: prettyprint(_, d+1, encoding), o)))
        elif o == None:
            return u'null'
        elif isinstance(o, bool):
            return unicode(str(o).lower())
        elif isinstance(o, dict):
            return prettyprint(o, d, encoding)
        elif isinstance(o, decimal.Decimal):
            return str(o)
        else:
            return repr(o)
    if not isinstance(obj, dict):
        return format(obj, depth)
    if isinstance(obj, TaggedValue):
        t = '@%s ' % tag(obj)
        obj = untagged(obj)
    else:
        t = ''
    r = [u'%s{\n' % t]
    values = []
    depth += 1
    for k, v in sorted(obj.iteritems()):
        space = u'    ' * depth
        val = format(v, depth)
        values.append(u'%s"%s":%s' % (space, k, val))
    r.append(',\n'.join(values))
    r.append(u'    ' * (depth - 1))
    r.append(u'\n%s}' % (u'    ' * (depth - 1)))
    return u''.join(r)

pp = prettyprint

import types
_tag_class_dict = {
    int: u'integer',
    decimal.Decimal: u'number',
    bool: u'boolean',
    dict: u'object',
    unicode: u'string',
    str: u'binary',
    types.NoneType: u'null',
    list: u'array',
    tuple: u'array',
}

_reserved = set(_tag_class_dict.values())

_builtin_types = dict([(b, [a]) for a, b in _tag_class_dict.items()])
_builtin_types['number'].append(int)

class TaggedValue(object):
    """type tag 付きのスカラー値用のクラス。
    {$tag:foo, $val:1} のようなオブジェクトは、あくまでもスカラー値として扱いたい。
    そのため、タグ付きのデータはすべて {} や dict などではなく
    専用のタグ型によって表現する。
    """
    def __init__(self, t, v):
        self.__tag = t
        self.__value = v

    @property
    def value(self):
        return self.__value

    @property
    def tag(self):
        return self.__tag

    def __str__(self):
        return u'@%s %s' % (self.tag, str(self.value))

    def __eq__(self, o):
        return self.tag == tag(o) and self.value == untagged(o)

class TagOnly(object):
    u"""タグのみを持ち、対応する値は存在しない値。
    JSON には {$tag: "...", $no-value:true} の形式でエンコードされる。
    """
    def __init__(self, tag):
        self.tag = tag

    @property
    def value(self):
        raise JsonError(u'タグのみの値です')

    def __eq__(self, o):
        return isinstance(o, TagOnly) and self.tag == tag(o)

def tagged(tagname, val):
    if tagname in _reserved:
        if not type(val) in _builtin_types[tagname]:
            raise JsonError(u'%s は予約語です' % tagname)
        else:
            return val
    r = TaggedValue(tagname, val)
    return r

def tag(val, explicit=False):
    try:
        return val.tag
    except:
        if explicit:
            raise JsonError(u'明示的なタグが指定されていません')
        t = type(val)
        try:
            return _tag_class_dict[t]
        except:
            raise JsonError(u'対応する JSON 型が存在しません： %s' % (str(t)))

def untagged(val, explicit=False):
    try:
        return val.value
    except:
        if explicit:
            raise JsonError(u'明示的なタグが指定されていません')
        return val

def split_tag(val):
    return tag(val), untagged(val)

def obj2path(obj):
    u"""辞書形式の JSON オブジェクトをパス形式に変換する。
    パス形式の JSON は深さが 1 の辞書となり、
    入れ子関係はすべてフラットにされてキーの値がパス情報となる。
    すなわち以下のオブジェクトは

    {
        "a": 1,
        "b": {
            "c": 2
        }
    }

    このような形式に変換される。

    {
        "a": 1,
        "b.c": 2
    }
    """
    assert isinstance(obj, dict)
    def flatten(obj, parent='$'):
        if isinstance(obj, dict):
            n = 0
            for k, v in obj.items():
                n += 1
                for r in flatten(v, parent + '.' + k):
                    yield r
            if n == 0: # 空のオブジェクト
                yield parent + '.' + '{}', None
        elif isinstance(obj, (list, tuple)):
            n = 0
            for i, v in enumerate(obj):
                n += 1
                for r in flatten(v, parent + '.' + str(i)):
                    yield r
            if n == 0:# 空のリスト
                yield parent + '.' + '[]', None
        else:
            yield parent, obj
    p = {}
    p.update(list(flatten(obj)))
    return p

from caty.util import merge_dict, try_parse
def path2obj(obj):
    u"""パス形式のオブジェクトを入れ子構造に戻す。
    """
    assert isinstance(obj, dict)
    def split_path(o):
        for k, v in o.items():
            k = k[2:] # $. の削除
            keys = k.split('.')
            if keys[-1] == '[]':
                keys.pop(-1)
                v = []
            elif keys[-1] == '{}':
                keys.pop(-1)
                v = {}
            yield reduce(lambda x, y: {y: x}, reversed(keys + [v]))
    
    def merge_path(o):
        return reduce(lambda a, b: merge_dict(a, b), split_path(o))

    def is_array(v):
        return try_parse(int, v.keys()[0]) != None
    
    def make_array(o):
        l = range(len(o))
        for k, v in o.items():
            l[int(k)] = make_tree(v)
        return l
    
    def make_obj(tree):
        o = {}
        for k, v in tree.items():
            o[k] = make_tree(v)
        return o
    
    def make_tree(o):
        if isinstance(o, dict):
            if is_array(o):
                return make_array(o)
            return make_obj(o)
        else:
            return o

    return make_tree(merge_path(obj))

import types
def is_json(obj):
    if isinstance(obj, (int, decimal.Decimal, unicode, str, bool, types.NoneType)):
        return True
    elif isinstance(obj, dict):
        return all(map(is_json, obj.values()))
    elif isinstance(obj, list):
        return all(map(is_json, obj))
    else:
        return False



