#coding:utf-8
class SortedList(list):
    u"""挿入した値が常にソートされているリスト。
    """
    def __init__(self, comparator=cmp):
        list.__init__(self)
        self.cmp = comparator

    def add(self, obj):
        self.append(obj)
        self.sort(cmp=self.cmp)

def fork(iterator, filter_func):
    u"""filter_func が True になるグループと False になるグループの二つにイテレータを分割する。
    戻り値の値は二要素のタプルであり、タプルの要素はリストである。

    >>> fork([0, 1, 2, 3, 4], lambda x: x % 2 == 0)
    ([0, 2, 4], [1, 3])
    """
    a = []
    b = []
    for i in iterator:
        if filter_func(i):
            a.append(i)
        else:
            b.append(i)
    return a, b

def merge_dict(a, b, conflict_mode='post'):
    u"""再帰的に辞書をマージしていく。
    conflict_mode は a, b でキーが重複する葉に対しての動作である。
    デフォルトでは 'post' が使われる。

    'post': b の値を使う
    'pre': a の値を使う
    'error': 例外を送出する

    関数オブジェクトを conflict_mode に渡すことも可能である。
    その関数は a[k] と b[k] の二つの値を引数として受け取れなければならない。
    関数の戻り値の値がマージ結果の値となる
    """
    def copy(value):
        if isinstance(value, dict):
            new = {}
            for k, v in value.items():
                new[k] = copy(v)
            return new
        elif isinstance(value, list):
            return value[:]
        else:
            return value
    n = {}
    for k, v in a.items():
        n[k] = copy(v)

    for k, v in b.items():
        if k in n:
            if isinstance(v, dict) and isinstance(n[k], dict):
                n[k] = merge_dict(n[k], v, conflict_mode)
            else:
                if conflict_mode == 'post':
                    n[k] = copy(v)
                elif conflict_mode == 'pre':
                    pass
                elif conflict_mode == 'error':
                    raise Exception('Conflict: %s' % str(k))
                elif callable(conflict_mode):
                    r = conflict_mode(n[k], v)
                    n[k] = copy(v)
                else:
                    raise Exception('Unknown mode: %s' % str(conflict_mode))
        else:
            n[k] = copy(v)
    return n

class MultiMap(object):
    def __init__(self):
        self.__dict = {}

    def __setitem__(self, k, v):
        if not k in self.__dict:
            self.__dict[k] = [v]
        else:
            self.__dict[k].append(v)

    def __contains__(self, k):
        return k in self.__dict

    def __getitem__(self, k):
        return self.__dict[k]

    def keys(self):
        return self.__dict.keys()

    def values(self):
        return self.__dict.values()

    def items(self):
        return self.__dict.items()

    def update(self, o):
        for k, v in o.items():
            for i in v:
                self[k] = i

    def __repr__(self):
        r = []
        for k, v in self.items():
            r.append('%s: %s' % (repr(k), repr(v)))
        return '{%s}' % ', '.join(r)

    __str__ = __repr__

import operator
class ImmutableDict(dict):
    u"""dict が unhashable で memoize できないので、代わりのオブジェクトを作る。
    """
    def __init__(self, o):
        p = {}
        for k, v in o.items():
            if isinstance(v, dict):
                p[k] = ImmutableDict(v)
            else:
                p[k] = v
        dict.__init__(self, p)
        self.__hash = self._make_hash()

    def __setitem__(self, k, v):
        raise Exception('Immutable')

    def _make_hash(self):
        r = []
        for k, v in self.items():
            r.append('%s%s' % (repr(k), repr(v)))
        return hash(reduce(operator.add, r))

    def __hash__(self):
        return self.__hash

